import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
import unittest

from metacontroller import Do
from metacontroller.internal.method_inspector import MethodInspector


class TestAssignments(unittest.TestCase):
    def test_controlled_method_assignments(self):

        def pre_controller_no_args(cls, arg1):
            cls.pre_controller_passed = arg1

        class BasicDo(Do):
            pre_controller_passed = False
            post_controller_passed = False
            pre_controller = pre_controller_no_args
            action = staticmethod(lambda: 1)
            post_controller = (
                lambda self, keyword_arg0=False: self.set_post_controller_value(
                    keyword_arg0
                )
            )

            def set_post_controller_value(self, value: bool):
                self.post_controller_passed = value

        inst = BasicDo()
        result = inst(True, keyword_arg0=True)
        self.assertTrue(inst.pre_controller_passed)
        self.assertTrue(result == 1)
        self.assertTrue(inst.post_controller_passed)


class DummyContextManager:
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass


def dummy_decorator(func):
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


class TestMethodReturns(unittest.TestCase):

    def test_value_return(self):
        def return_explicit_value():
            return 42

        fn = MethodInspector(return_explicit_value)
        self.assertTrue(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_conditional_value_return(self):
        def return_in_conditional(x):
            if x > 0:
                return x
            else:
                return -x

        fn = MethodInspector(return_in_conditional)
        self.assertTrue(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_loop_value_return(self):
        def return_in_loop(lst):
            for item in lst:
                if item > 0:
                    return item
            return

        fn = MethodInspector(return_in_loop)
        self.assertTrue(fn.has_explicit_value_return)
        self.assertTrue(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_try_except_value_return(self):
        def return_in_try_except(x):
            try:
                return 1 / x
            except ZeroDivisionError:
                return float("inf")

        fn = MethodInspector(return_in_try_except)
        self.assertTrue(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_nested_function_value_return(self):
        def outer_function():
            def inner_function():
                return "inner"

            return inner_function()

        fn = MethodInspector(outer_function)
        self.assertTrue(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_lambda_function_value_return(self):
        def return_from_lambda():
            f = lambda x: x + 1
            return f(5)

        fn = MethodInspector(return_from_lambda)
        self.assertTrue(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_no_return(self):
        def no_return():
            pass

        fn = MethodInspector(no_return)
        self.assertFalse(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_return_none(self):
        def return_none():
            return

        fn = MethodInspector(return_none)
        self.assertFalse(fn.has_explicit_value_return)
        self.assertTrue(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_no_return_in_conditional(self):
        def no_return_in_conditional(x):
            if x > 0:
                x += 1
            else:
                x -= 1

        fn = MethodInspector(no_return_in_conditional)
        self.assertFalse(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_no_return_in_loop(self):
        def no_return_in_loop(lst):
            for item in lst:
                item += 1

        fn = MethodInspector(no_return_in_loop)
        self.assertFalse(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_no_return_in_try_except(self):
        def no_return_in_try_except(x):
            try:
                x += 1
            except:
                x -= 1

        fn = MethodInspector(no_return_in_try_except)
        self.assertFalse(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_no_return_nested_function(self):
        def no_return_outer():
            def no_return_inner():
                pass

            no_return_inner()

        fn = MethodInspector(no_return_outer)
        self.assertFalse(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_no_return_nested_function_with_inner_return(self):
        def no_return_outer():
            def no_return_inner():
                return True

            no_return_inner()

        fn = MethodInspector(no_return_outer)
        self.assertFalse(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_mixed_return(self):
        def mixed_return(x):
            if x > 0:
                return x
            x -= 1

        fn = MethodInspector(mixed_return)
        self.assertTrue(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_return_in_loop_no_after(self):
        def return_in_loop_no_after(lst):
            for item in lst:
                if item > 0:
                    return item
            print("No positive numbers")

        fn = MethodInspector(return_in_loop_no_after)
        self.assertTrue(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_mixed_try_except(self):
        def mixed_try_except(x):
            try:
                return 1 / x
            except ZeroDivisionError:
                x -= 1

        fn = MethodInspector(mixed_try_except)
        self.assertTrue(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_yield_value(self):
        def generator_function():
            yield 42

        fn = MethodInspector(generator_function)
        self.assertFalse(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertTrue(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_yield_in_loop(self):
        def generator_in_loop(lst):
            for item in lst:
                yield item

        fn = MethodInspector(generator_in_loop)
        self.assertFalse(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertTrue(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_yield_in_conditional(self):
        def generator_in_conditional(x):
            if x > 0:
                yield x
            else:
                yield -x

        fn = MethodInspector(generator_in_conditional)
        self.assertFalse(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertTrue(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_yield_from_value(self):
        def generator_yield_from():
            yield from [1, 2, 3]

        fn = MethodInspector(generator_yield_from)
        self.assertFalse(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertTrue(fn.has_value_yield_from)

    def test_yield_from_in_loop(self):
        def generator_yield_from_loop(lst):
            for sublist in lst:
                yield from sublist

        fn = MethodInspector(generator_yield_from_loop)
        self.assertFalse(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertTrue(fn.has_value_yield_from)

    def test_yield_and_return(self):
        def generator_with_return(x):
            yield x
            return x

        fn = MethodInspector(generator_with_return)
        self.assertTrue(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertTrue(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)


class TestFunctionsWithClasses(unittest.TestCase):

    def test_function_with_inner_class_return(self):
        def function_with_class():
            class InnerClass:
                def method(self):
                    return 42

            return "outer"

        fn = MethodInspector(function_with_class)
        self.assertTrue(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_function_with_inner_class_no_return(self):
        def function_with_class():
            class InnerClass:
                def method(self):
                    return 42

            pass

        fn = MethodInspector(function_with_class)
        self.assertFalse(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_function_with_inner_class_and_with_block(self):
        def function_with_class_and_with():
            class InnerClass:
                def method(self):
                    return 42

            with DummyContextManager():
                pass
            return "outer"

        fn = MethodInspector(function_with_class_and_with)
        self.assertTrue(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_function_with_inner_class_and_closure(self):
        def function_with_class_and_closure():
            class InnerClass:
                def method(self):
                    return 42

            def inner_function():
                return "inner"

            return inner_function()

        fn = MethodInspector(function_with_class_and_closure)
        self.assertTrue(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_function_with_inner_class_and_lambda(self):
        def function_with_class_and_lambda():
            class InnerClass:
                def method(self):
                    return 42

            f = lambda x: x + 1
            return f(5)

        fn = MethodInspector(function_with_class_and_lambda)
        self.assertTrue(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_function_with_inner_class_yield(self):
        def function_with_class_and_yield():
            class InnerClass:
                def method(self):
                    yield 42

            return "outer"

        fn = MethodInspector(function_with_class_and_yield)
        self.assertTrue(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)

    def test_function_with_inner_class_yield_from(self):
        def function_with_class_and_yield_from():
            class InnerClass:
                def method(self):
                    yield from [1, 2, 3]

            return "outer"

        fn = MethodInspector(function_with_class_and_yield_from)
        self.assertTrue(fn.has_explicit_value_return)
        self.assertFalse(fn.has_explicit_void_return)
        self.assertFalse(fn.has_value_yield)
        self.assertFalse(fn.has_value_yield_from)


class TestMethodInspectorAdditionalProperties(unittest.TestCase):

    def test_decorated_function(self):
        @dummy_decorator
        def decorated_function():
            return 42

        fn = MethodInspector(decorated_function)
        self.assertFalse(fn.is_staticmethod)
        self.assertFalse(fn.is_lambda)

    def test_lambda_function(self):
        lambda_function = lambda x: x + 1

        fn = MethodInspector(lambda_function)
        self.assertFalse(fn.is_staticmethod)
        self.assertTrue(fn.is_lambda)

    def test_decorated_lambda_function(self):
        decorated_lambda = dummy_decorator(lambda x: x + 1)

        fn = MethodInspector(decorated_lambda)
        self.assertFalse(fn.is_staticmethod)
        self.assertFalse(fn.is_lambda)

    def test_function_with_inner_class_and_decorator(self):
        def function_with_class_and_decorator():
            @dummy_decorator
            class InnerClass:
                def method(self):
                    return 42

            return "outer"

        fn = MethodInspector(function_with_class_and_decorator)
        self.assertFalse(fn.is_staticmethod)
        self.assertFalse(fn.is_lambda)


class ArgClass:
    # Example functions for testing
    def simple_function(self, a, /, b=10, *args, c, d=20, **kwargs):
        pass

    def function_with_annotations(self, x: int, y: str = "default") -> float:
        pass


class TestMethodInspectorProperties(unittest.TestCase):

    def setUp(self):
        # Setup MethodInspector instances for each function
        inst = ArgClass()
        self.simple_fn = MethodInspector(inst.simple_function)
        self.annotation_fn = MethodInspector(inst.function_with_annotations)

    def test_posonlyargs(self):
        self.assertEqual(self.simple_fn.posonlyargs, ["a"])

    def test_args(self):
        self.assertEqual(self.simple_fn.args, ["self", "a", "b"])

    def test_call_args(self):
        self.assertEqual(self.simple_fn.call_args, ["a", "b"])

    def test_num_parameters(self):
        self.assertEqual(self.simple_fn.num_parameters, 7)

    def test_num_call_parameters(self):
        self.assertEqual(self.simple_fn.num_call_parameters, 6)

    def test_varargs(self):
        self.assertEqual(self.simple_fn.varargs, "args")

    def test_varkw(self):
        self.assertEqual(self.simple_fn.varkw, "kwargs")

    def test_defaults(self):
        self.assertEqual(self.simple_fn.defaults, (10,))

    def test_kwonlyargs(self):
        self.assertEqual(self.simple_fn.kwonlyargs, ["c", "d"])

    def test_kwonlydefaults(self):
        self.assertEqual(self.simple_fn.kwonlydefaults, {"d": 20})

    def test_annotations(self):
        self.assertEqual(
            self.annotation_fn.annotations, {"x": int, "y": str, "return": float}
        )

    def test_has_arg_unpack(self):
        self.assertTrue(self.simple_fn.has_arg_unpack)

    def test_has_kwarg_unpack(self):
        self.assertTrue(self.simple_fn.has_kwarg_unpack)


if __name__ == "__main__":
    unittest.main()



####
# Arguments
CHOSEN_ARG_NAME = "__ctrl_chosen__"
PARTITION_ARG_NAME = "__ctrl_partition__"
K_ARG_NAME = "__ctrl_k__"
SORT_CMP_ARG_A_NAME = "__ctrl_a__"
SORT_CMP_ARG_B_NAME = "__ctrl_b__"
CLASS_ARG_NAME = "self"


####
# Methods
PRE_CONTROLLER_METHOD_NAME = "pre_controller"
FILTER_METHOD_NAME = "filter"
SORT_KEY_METHOD_NAME = "sort_key"
SORT_CMP_METHOD_NAME = "sort_cmp"
ACTION_METHOD_NAME = "action"
FOLD_METHOD_NAME = "fold"
POST_CONTROLLER_METHOD_NAME = "post_controller"
GENERATED_CALL_METHOD_NAME = "__ctrl_call__"


####
# Variable Names
ACTION_RESULT_ASSIGNMENT_NAME = "__ctrl_result__"

RESERVED_KEYWORDS = {
    CHOSEN_ARG_NAME,
    PARTITION_ARG_NAME,
    K_ARG_NAME,
    SORT_CMP_ARG_A_NAME,
    SORT_CMP_ARG_B_NAME,
    ACTION_RESULT_ASSIGNMENT_NAME,
}



import ast
from functools import cmp_to_key
from heapq import nlargest, nsmallest
from itertools import islice
from typing import Any, Callable

from metacontroller.internal.exceptions import (
    InvalidControllerMethodError,
    InvalidReturnError,
)
from metacontroller.internal.method_invocation import MethodInvocation
from metacontroller.internal.namespace import (
    ACTION_METHOD_NAME,
    ACTION_RESULT_ASSIGNMENT_NAME,
    CLASS_ARG_NAME,
    FILTER_METHOD_NAME,
    FOLD_METHOD_NAME,
    GENERATED_CALL_METHOD_NAME,
    K_ARG_NAME,
    PARTITION_ARG_NAME,
    POST_CONTROLLER_METHOD_NAME,
    PRE_CONTROLLER_METHOD_NAME,
    SORT_CMP_METHOD_NAME,
    SORT_KEY_METHOD_NAME,
)

from ._base import BaseControllerImplementation


class DoKImplementation(BaseControllerImplementation):
    def __init__(self, cls, name, bases, attrs, stack_frame) -> None:
        super().__init__(cls, name, bases, attrs, stack_frame)

    def validate(self) -> None:
        super().validate()
        if self.has_sort_key and self.has_sort_cmp:
            err = f'DoK controller "{self.name}" is invalid because both sort methods ("{SORT_KEY_METHOD_NAME}" and "{SORT_CMP_METHOD_NAME}") are defined.'
            err += f' You must define only one. Note that "{SORT_KEY_METHOD_NAME}" is more performant.'
            raise InvalidControllerMethodError(err)

        if self.has_filter:
            if len(self.filter.call_args) < 1:
                raise AttributeError(
                    f'"{FILTER_METHOD_NAME}" should be defined with at least 1 non-class argument (chosen), but 0 were given.'
                )

        if self.has_sort_key:
            if len(self.sort_key.call_args) < 1:
                raise AttributeError(
                    f'"{SORT_KEY_METHOD_NAME}" should be defined with at least 1 non-class argument (chosen), but 0 were given.'
                )

        if self.has_sort_cmp:
            if len(self.sort_cmp.call_args) < 2:
                raise AttributeError(
                    f'"{SORT_CMP_METHOD_NAME}" should be defined with at least 2 non-class arguments (a, b), but {len(self.sort_cmp.call_args)} were given.'
                )

        if self.has_action:
            if len(self.action.call_args) < 1:
                raise AttributeError(
                    f'"{ACTION_METHOD_NAME}" should be defined with at least 1 non-class argument (chosen), but 0 were given.'
                )

        if self.has_fold:
            if len(self.fold.call_args) < 1:
                raise AttributeError(
                    f'"{FOLD_METHOD_NAME}" should be defined with at least 1 non-class argument (list of action results), but 0 were given.'
                )

            if self.has_action and not self.action.returns_a_value:
                raise InvalidReturnError(
                    f'"{FOLD_METHOD_NAME}" was defined, but "{ACTION_METHOD_NAME}" does not return anything.'
                )

    def generate_call_method(self) -> Callable[..., Any]:
        body = []
        additional_globals = {}
        get_elements = ast.Name(id=PARTITION_ARG_NAME, ctx=ast.Load())

        if self.has_pre_controller:
            pre_controller_call = MethodInvocation(
                self.pre_controller
            ).to_function_call(name=PRE_CONTROLLER_METHOD_NAME)
            body.append(ast.Expr(value=pre_controller_call))

        if self.has_filter:
            if self.filter.num_call_parameters != 1:
                filter_fn = MethodInvocation(self.filter).to_lambda(
                    [self.filter.call_args[0]], name=FILTER_METHOD_NAME
                )
            else:
                filter_fn = ast.Attribute(
                    value=ast.Name(id=CLASS_ARG_NAME, ctx=ast.Load()),
                    attr=FILTER_METHOD_NAME,
                    ctx=ast.Load(),
                )
            get_elements = ast.Call(
                func=ast.Name(id="filter", ctx=ast.Load()),
                args=[filter_fn, get_elements],
                keywords=[],
            )

        if self.has_sort_key:
            if self.sort_key.num_call_parameters != 1:
                sort_fn_key = MethodInvocation(self.sort_key).to_lambda(
                    [self.sort_key.call_args[0]], name=SORT_KEY_METHOD_NAME
                )
            else:
                sort_fn_key = ast.Attribute(
                    value=ast.Name(id=CLASS_ARG_NAME, ctx=ast.Load()),
                    attr=SORT_KEY_METHOD_NAME,
                    ctx=ast.Load(),
                )

            if self.cls.reverse_sort:
                sort_fn = ast.Name(id="nlargest", ctx=ast.Load())
                additional_globals["nlargest"] = nlargest
            else:
                sort_fn = ast.Name(id="nsmallest", ctx=ast.Load())
                additional_globals["nsmallest"] = nsmallest

            get_elements = ast.Call(
                func=sort_fn,
                args=[ast.Name(id=K_ARG_NAME, ctx=ast.Load()), get_elements],
                keywords=[ast.keyword(arg="key", value=sort_fn_key)],
            )

        if self.has_sort_cmp:
            if self.sort_cmp.num_call_parameters != 2:
                sort_fn_key = MethodInvocation(self.sort_cmp).to_lambda(
                    self.sort_cmp.call_args[:2], name=SORT_CMP_METHOD_NAME
                )
            else:
                sort_fn_key = ast.Attribute(
                    value=ast.Name(id=CLASS_ARG_NAME, ctx=ast.Load()),
                    attr=SORT_CMP_METHOD_NAME,
                    ctx=ast.Load(),
                )

            if self.cls.reverse_sort:
                sort_fn = ast.Name(id="nlargest", ctx=ast.Load())
                additional_globals["nlargest"] = nlargest
            else:
                sort_fn = ast.Name(id="nsmallest", ctx=ast.Load())
                additional_globals["nsmallest"] = nsmallest

            get_elements = ast.Call(
                func=sort_fn,
                args=[ast.Name(id=K_ARG_NAME, ctx=ast.Load()), get_elements],
                keywords=[
                    ast.keyword(
                        arg="key",
                        value=ast.Call(
                            func=ast.Name(id="cmp_to_key", ctx=ast.Load()),
                            args=[sort_fn_key],
                            keywords=[],
                        ),
                    )
                ],
            )
            additional_globals["cmp_to_key"] = cmp_to_key

        if not self.has_sort_key and not self.has_sort_cmp:
            get_elements = ast.Call(
                func=ast.Name(id="islice", ctx=ast.Load()),
                args=[get_elements, ast.Name(id=K_ARG_NAME, ctx=ast.Load())],
                keywords=[],
            )
            additional_globals["islice"] = islice
            if not self.has_action:
                get_elements = ast.Call(
                    func=ast.Name(id="list", ctx=ast.Load()),
                    args=[get_elements],
                    keywords=[],
                )

        if self.has_action:
            action_invoke = MethodInvocation(self.action)
            action_args, action_keywords = action_invoke.get_call_args_and_keywords()

            if self.action.returns_a_value:
                # we should capture the results using map
                if self.action.num_call_parameters != 1:
                    # action has additional parameters, use a lambda
                    action_fn = action_invoke.to_lambda(
                        [action_args[0].id], name=ACTION_METHOD_NAME
                    )
                else:
                    # action only takes the required chosen parameter
                    action_fn = ast.Attribute(
                        value=ast.Name(id=CLASS_ARG_NAME, ctx=ast.Load()),
                        attr=ACTION_METHOD_NAME,
                        ctx=ast.Load(),
                    )

                action_call = ast.Call(
                    func=ast.Name(id="list", ctx=ast.Load()),
                    args=[
                        ast.Call(
                            func=ast.Name(id="map", ctx=ast.Load()),
                            args=[action_fn, get_elements],
                            keywords=[],
                        )
                    ],
                    keywords=[],
                )
                action = ast.Assign(
                    targets=[
                        ast.Name(id=ACTION_RESULT_ASSIGNMENT_NAME, ctx=ast.Store())
                    ],
                    value=action_call,
                )

            else:
                # no need to capture the result from the action, so use a basic for loop
                action = ast.For(
                    target=ast.Name(id=action_args[0].id, ctx=ast.Store()),
                    iter=get_elements,
                    body=[
                        ast.Expr(
                            value=action_invoke.to_function_call(
                                action_args, action_keywords, name=ACTION_METHOD_NAME
                            )
                        )
                    ],
                    orelse=[],
                )

            body.append(action)

        if self.has_fold:
            fold_invoke = MethodInvocation(self.fold)
            fold_args, fold_keywords = fold_invoke.get_call_args_and_keywords()
            fold_args.pop(0)

            if self.has_action:
                fold_args.insert(
                    0, ast.Name(id=ACTION_RESULT_ASSIGNMENT_NAME, ctx=ast.Load())
                )
            else:
                fold_args.insert(0, get_elements)

            fold_assignment = ast.Assign(
                targets=[ast.Name(id=ACTION_RESULT_ASSIGNMENT_NAME, ctx=ast.Store())],
                value=fold_invoke.to_function_call(
                    fold_args, fold_keywords, name=FOLD_METHOD_NAME
                ),
            )
            body.append(fold_assignment)

        elif not self.has_action:
            # does not have an action, return whatever is get_elements
            get_elements_result = ast.Assign(
                targets=[ast.Name(id=ACTION_RESULT_ASSIGNMENT_NAME, ctx=ast.Store())],
                value=get_elements,
            )
            body.append(get_elements_result)

        if self.has_post_controller:
            post_controller_call = MethodInvocation(
                self.post_controller
            ).to_function_call(name=POST_CONTROLLER_METHOD_NAME)
            body.append(ast.Expr(value=post_controller_call))

        if not self.has_fold and (self.has_action and not self.action.returns_a_value):
            pass  # do nothing since we explicitly do not need a return value here
        else:
            body.append(
                ast.Return(
                    value=ast.Name(id=ACTION_RESULT_ASSIGNMENT_NAME, ctx=ast.Load())
                )
            )

        args, saved_defaults = self.get_call_args(
            use_class_arg=True,
            use_k_arg=True,
            use_partition_arg=True,
        )
        additional_globals.update(saved_defaults)
        call_fn = ast.FunctionDef(
            name=GENERATED_CALL_METHOD_NAME,
            args=args,
            body=body,
            decorator_list=[],
            type_params=[],
        )

        module = ast.fix_missing_locations(ast.Module(body=[call_fn], type_ignores=[]))
        return self.compile_call_method(module, additional_globals)



import inspect
from typing import Any, Generic, Iterable, List, Protocol, TypeVar, Union, _ProtocolMeta

from .classes.do import DoImplementation
from .classes.do_all import DoAllImplementation
from .classes.do_k import DoKImplementation
from .classes.do_one import DoOneImplementation

TChosen = TypeVar("TChosen")
TActionReturn = TypeVar("TActionReturn")
TFoldReturn = TypeVar("TFoldReturn")
TSupportsRichComparison = TypeVar(
    "TSupportsRichComparison", bound="SupportsRichComparison"
)


class SupportsRichComparison(Protocol):
    def __lt__(self, other: TSupportsRichComparison) -> bool: ...
    def __gt__(self, other: TSupportsRichComparison) -> bool: ...


class MetaController(_ProtocolMeta):

    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)
        if name in ["Do", "DoOne", "DoK", "DoAll"]:
            return

        _base_classes = [base for base in bases if base in [Do, DoOne, DoK, DoAll]]
        if len(_base_classes) > 1:
            raise TypeError("Controller multiple inheritance is not allowed.")
        _base_class = _base_classes[0]

        if _base_class is Do:
            implementation = DoImplementation
        elif _base_class is DoOne:
            implementation = DoOneImplementation
        elif _base_class is DoK:
            implementation = DoKImplementation
        elif _base_class is DoAll:
            implementation = DoAllImplementation
        else:
            raise NotImplementedError("Unkown base class.")  # should not get here

        controller = implementation(
            cls, name, bases, attrs, inspect.currentframe().f_back
        )
        controller.validate()
        cls.__call__ = controller.generate_call_method()


class Do(Generic[TActionReturn], metaclass=MetaController):
    optimize: bool = False

    ###
    # Valid User Defined Methods:
    #

    def pre_controller(self) -> None: ...

    def action(self) -> TActionReturn: ...

    def post_controller(self) -> None: ...

    ###
    # Built-in Instance Methods
    #

    def __call__(self, *args: Any, **kwargs: Any) -> Union[TActionReturn, None]:
        """Generated call method for your controller. *args and **kwargs represent the
        position only, arguments, defaulted arguments, argument unpacks, keyword only,
        and keyword unpacks that you have defined (if any) across all your controlled
        methods.

        Returns:
            TActionReturn: If your action returns a value, that will be the return, else
            this method will return None.
        """
        ...

    def update_to(self, cls: "Do[TActionReturn]") -> None: ...


class DoOne(Generic[TChosen, TActionReturn], metaclass=MetaController):
    optimize: bool = False
    reverse_sort: bool = False

    ###
    # Valid User Defined Methods:
    #

    def pre_controller(self) -> None: ...

    def filter(self, chosen: TChosen) -> bool: ...

    def sort_key(self, chosen: TChosen) -> SupportsRichComparison: ...

    def sort_cmp(self, a: TChosen, b: TChosen) -> int: ...

    def action(self, chosen: TChosen) -> TActionReturn: ...

    def post_controller(self) -> None: ...

    ###
    # Built-in Instance Methods
    #

    def __call__(
        self, partition: Iterable[TChosen], /, *args: Any, **kwargs: Any
    ) -> Union[TActionReturn, None]:
        """Generated call method for your controller. *args and **kwargs represent the
        position only, arguments, defaulted arguments, argument unpacks, keyword only,
        and keyword unpacks that you have defined (if any) across all your controlled
        methods.

        Args:
            partition (Iterable[TChosen]): Iterable representing the set of elements
            this controller will operate over.

        Returns:
            TActionReturn: If your action returns a value, that will be the return, else
            this method will return None.
        """
        ...

    def update_to(
        self,
        cls: Union[
            "DoOne[TChosen, TActionReturn]",
            "DoAll[TChosen, TActionReturn, TActionReturn]",
        ],
    ) -> None: ...


class DoK(Generic[TChosen, TActionReturn, TFoldReturn], metaclass=MetaController):
    optimize: bool = False
    reverse_sort: bool = False

    ###
    # Valid User Defined Methods:
    #

    def pre_controller(self) -> None: ...

    def filter(self, chosen: TChosen) -> bool: ...

    def sort_key(self, chosen: TChosen) -> SupportsRichComparison: ...

    def sort_cmp(self, a: TChosen, b: TChosen) -> int: ...

    def action(self, chosen: TChosen) -> TActionReturn: ...

    def post_controller(self) -> None: ...

    def fold(self, results: List[TActionReturn]) -> TFoldReturn: ...

    ###
    # Built-in Instance Methods
    #

    def __call__(
        self, k: int, partition: Iterable[TChosen], /, *args: Any, **kwargs: Any
    ) -> Union[Iterable[TActionReturn], TFoldReturn, None]:
        """Generated call method for your controller. *args and **kwargs represent the
        position only, arguments, defaulted arguments, argument unpacks, keyword only,
        and keyword unpacks that you have defined (if any) across all your controlled
        methods.

        Args:
            k (int): Number of chosen elements to operate over from the partition.
            partition (Iterable[TChosen]): Iterable representing the set of elements
            this controller will operate over.

        Returns:
            Union[Iterable[TActionReturn], TFoldReturn, None]: If action returns a value and there
            is no fold(...) method defined, the result will be an Iterable of elements
            returned from k number of calls to the action(...). If fold(...) is defined,
            this will return the result from the fold(...) method. Else, this will return
            None.
        """
        ...

    def update_to(self, cls: "DoK[TChosen, TActionReturn, TFoldReturn]") -> None: ...


class DoAll(Generic[TChosen, TActionReturn, TFoldReturn], metaclass=MetaController):
    optimize: bool = False
    reverse_sort: bool = False

    ###
    # Valid User Defined Methods:
    #

    def pre_controller(self) -> None: ...

    def filter(self, chosen: TChosen) -> bool: ...

    def sort_key(self, chosen: TChosen) -> SupportsRichComparison: ...

    def sort_cmp(self, a: TChosen, b: TChosen) -> int: ...

    def action(self, chosen: TChosen) -> TActionReturn: ...

    def post_controller(self) -> None: ...

    def fold(self, results: List[TActionReturn]) -> TFoldReturn: ...

    ###
    # Built-in Instance Methods
    #

    def __call__(
        self, partition: Iterable[TChosen], /, *args: Any, **kwargs: Any
    ) -> Union[Iterable[TActionReturn], TFoldReturn, None]:
        """Generated call method for your controller. *args and **kwargs represent the
        position only, arguments, defaulted arguments, argument unpacks, keyword only,
        and keyword unpacks that you have defined (if any) across all your controlled
        methods.

        Args:
            partition (Iterable[TChosen]): Iterable representing the set of elements
            this controller will operate over.

        Returns:
            Union[Iterable[TActionReturn], TFoldReturn, None]: If action returns a value and there
            is no fold(...) method defined, the result will be an Iterable of elements
            returned from all the calls to the action(...) method. If fold(...) is defined,
            this will return the result from the fold(...) method. Else, this will return
            None.
        """
        ...

    def update_to(
        self,
        cls: Union[
            "DoAll[TChosen, TActionReturn, TFoldReturn]",
            "DoOne[TChosen, TActionReturn, TFoldReturn]",
        ],
    ) -> None: ...



import ast
from typing import List, Tuple

from metacontroller.internal.method_inspector import MethodInspector
from metacontroller.internal.namespace import CLASS_ARG_NAME


class MethodInvocation:
    def __init__(self, method: MethodInspector) -> None:
        self.method = method

    def get_call_args_and_keywords(self) -> Tuple[List[ast.AST], List[ast.keyword]]:
        """
        Generate the call args and keywords for the instance method.

        Returns:
            Tuple[List[ast.AST], List[ast.keyword]]: Tuple of the list of arguments and the list of keywords
            to call this method.
        """
        args = [ast.Name(id=arg, ctx=ast.Load()) for arg in self.method.call_args]
        if self.method.has_arg_unpack:
            args.append(
                ast.Starred(
                    value=ast.Name(id=self.method.varargs, ctx=ast.Load()),
                    ctx=ast.Load(),
                )
            )

        keywords = [
            ast.keyword(arg=keyword, value=ast.Name(id=keyword, ctx=ast.Load()))
            for keyword, _ in self.method.get_keyword_only_args()
        ]

        if self.method.has_kwarg_unpack:
            keywords.append(
                ast.keyword(
                    arg=None, value=ast.Name(id=self.method.varkw, ctx=ast.Load())
                )
            )
        return args, keywords

    def to_function_call(
        self,
        args: List[ast.AST] = None,
        keywords: List[ast.keyword] = None,
        name: str = None,
    ) -> ast.Call:
        """
        Generates the invocation call for this method, assuming the arguments
        being passed in are the same name as the parameters defined in this
        function.

        Returns:
            ast.Call: ast representation of a Callable that will call this instances method.
            This should be wrapped in an ast.Expr before compiling.
        """

        if args is None or keywords is None:
            __args, __keywords = self.get_call_args_and_keywords()
            if args is None:
                args = __args
            if keywords is None:
                keywords = __keywords

        if name is None:
            _name = self.method.name
        else:
            _name = name
        return ast.Call(
            func=ast.Attribute(
                value=ast.Name(id=CLASS_ARG_NAME, ctx=ast.Load()),
                attr=_name,
                ctx=ast.Load(),
            ),
            args=args,
            keywords=keywords,
        )

    def to_lambda(self, lambda_args: List[str], name: str = None) -> ast.Lambda:
        args = ast.arguments(
            posonlyargs=[],
            args=[ast.arg(arg=arg, annotation=None) for arg in lambda_args],
            vararg=None,
            kwonlyargs=[],
            kw_defaults=[],
            kwarg=None,
            defaults=[],
        )
        return ast.Lambda(args=args, body=self.to_function_call(name=name))



import random
from typing import Any, List

random.seed(0)
import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
import unittest

from metacontroller import DoAll


class TestDoAllMethodCombinations(unittest.TestCase):
    def setUp(test_self):
        test_self.elements = [random.randint(0, 1000) for _ in range(10)]
        test_self.k = 10

    def test_filter_sort_cmp(test_self):
        class T(DoAll):
            def filter(self, chosen) -> bool:
                return chosen % 2 == 0

            def sort_cmp(self, a, b):
                return -1 if a < b else 1 if a > b else 0

        inst = T()
        expected_result = sorted([i for i in test_self.elements if i % 2 == 0])[
            : test_self.k
        ]
        result = inst(test_self.elements)
        test_self.assertListEqual(
            result,
            expected_result,
        )

    def test_filter_sort_key(test_self):
        class T(DoAll):
            def filter(self, chosen) -> bool:
                return chosen % 2 == 0

            def sort_key(self, chosen):
                return chosen

        inst = T()
        result = inst(test_self.elements)
        test_self.assertTrue(
            result
            == sorted(list(filter(lambda x: x % 2 == 0, test_self.elements)))[
                : test_self.k
            ]
        )

    def test_filter_action(test_self):
        class T(DoAll):
            def filter(self, chosen: Any) -> bool:
                return chosen % 2 == 0

            def action(self, chosen: Any) -> Any:
                return chosen + 1

        inst = T()
        result = inst(test_self.elements)
        test_self.assertTrue(
            result == [i + 1 for i in test_self.elements if i % 2 == 0][: test_self.k]
        )

    def test_filter_fold(test_self):
        class T(DoAll):
            def filter(self, chosen: Any) -> bool:
                return chosen % 2 == 0

            def fold(self, results: List) -> Any:
                return sum(results)

        inst = T()
        result = inst(test_self.elements)
        test_self.assertTrue(
            result
            == sum(
                list(filter(lambda x: x % 2 == 0, test_self.elements))[: test_self.k]
            )
        )

    def test_sort_cmp_action(test_self):
        class T(DoAll):
            def sort_cmp(self, a: Any, b: Any) -> int:
                return -1 if a < b else 1 if a > b else 0

            def action(self, chosen: Any) -> Any:
                return chosen + 1

        inst = T()
        result = inst(test_self.elements)
        expected_result = sorted([i + 1 for i in test_self.elements])[: test_self.k]
        test_self.assertListEqual(result, expected_result)

    def test_sort_cmp_fold(test_self):
        class T(DoAll):
            def sort_cmp(self, a: Any, b: Any) -> int:
                return -1 if a < b else 1 if a > b else 0

            def fold(self, results: List) -> Any:
                test_self.assertListEqual(
                    results, sorted(test_self.elements)[: test_self.k]
                )
                return sum(results)

        inst = T()
        result = inst(test_self.elements)
        test_self.assertTrue(result == sum(sorted(test_self.elements)[: test_self.k]))

    def test_sort_key_action(test_self):
        class T(DoAll):
            def sort_key(self, chosen: Any):
                return chosen

            def action(self, chosen):
                return chosen + 1

        inst = T()
        result = inst(test_self.elements)
        test_self.assertListEqual(
            result, sorted([i + 1 for i in sorted(test_self.elements)][: test_self.k])
        )

    def test_sort_key_fold(test_self):
        class T(DoAll):
            def sort_key(self, chosen: Any):
                return chosen

            def fold(self, results: List) -> Any:
                test_self.assertListEqual(
                    results, sorted(test_self.elements)[: test_self.k]
                )
                return sum(results)

        inst = T()
        result = inst(test_self.elements)
        test_self.assertTrue(result == sum(sorted(test_self.elements)[: test_self.k]))

    def test_action_fold(test_self):
        class T(DoAll):
            def action(self, chosen: Any) -> Any:
                return chosen + 1

            def fold(self, results: List) -> Any:
                test_self.assertListEqual(
                    results, [i + 1 for i in test_self.elements[: test_self.k]]
                )
                return sum(results)

        inst = T()
        result = inst(test_self.elements)
        test_self.assertTrue(
            result == sum([i + 1 for i in test_self.elements][: test_self.k])
        )

    def test_filter_sort_cmp_action(test_self):
        class T(DoAll):
            def filter(self, chosen: Any) -> bool:
                return chosen % 2 == 0

            def sort_cmp(self, a: Any, b: Any) -> int:
                return -1 if a < b else 1 if a > b else 0

            def action(self, chosen):
                return chosen + 1

        inst = T()
        result = inst(test_self.elements)
        expected_results = sorted(
            [i + 1 for i in test_self.elements if i % 2 == 0][: test_self.k]
        )
        test_self.assertListEqual(result, expected_results)

    def test_filter_sort_cmp_fold(test_self):
        class T(DoAll):
            def filter(self, chosen: Any) -> bool:
                return chosen % 2 == 0

            def sort_cmp(self, a: Any, b: Any) -> int:
                return -1 if a < b else 1 if a > b else 0

            def fold(self, results: List) -> Any:
                test_self.assertListEqual(
                    results,
                    sorted([i for i in test_self.elements if i % 2 == 0])[
                        : test_self.k
                    ],
                )
                return sum(results)

        inst = T()
        result = inst(test_self.elements)
        expected_results = sum(
            sorted([i for i in test_self.elements if i % 2 == 0])[: test_self.k]
        )
        test_self.assertTrue(result == expected_results)

    def test_filter_sort_key_action(test_self):
        class T(DoAll):
            def filter(self, chosen: Any) -> bool:
                return chosen % 2 == 0

            def sort_key(self, chosen: Any):
                return chosen

            def action(self, chosen: Any) -> Any:
                return chosen + 1

        inst = T()
        result = inst(test_self.elements)
        expected_result = sorted([i + 1 for i in test_self.elements if i % 2 == 0])[
            : test_self.k
        ]
        test_self.assertListEqual(result, expected_result)

    def test_filter_sort_key_fold(test_self):
        class T(DoAll):
            def filter(self, chosen: Any) -> bool:
                return chosen % 2 == 0

            def sort_key(self, chosen: Any):
                return chosen

            def fold(self, results: List) -> Any:
                test_self.assertListEqual(
                    results,
                    sorted([i for i in test_self.elements if i % 2 == 0])[
                        : test_self.k
                    ],
                )
                return sum(results)

        inst = T()
        result = inst(test_self.elements)
        expected_results = sum(
            sorted([i for i in test_self.elements if i % 2 == 0])[: test_self.k]
        )
        test_self.assertTrue(result == expected_results)

    def test_filter_action_fold(test_self):
        class T(DoAll):
            def filter(self, chosen: Any) -> bool:
                return chosen % 2 == 0

            def action(self, chosen: Any) -> Any:
                return chosen + 1

            def fold(self, results: List) -> Any:
                test_self.assertListEqual(
                    results,
                    [i + 1 for i in test_self.elements if i % 2 == 0][: test_self.k],
                )
                return sum(results)

        inst = T()
        result = inst(test_self.elements)
        expected_results = sum(
            [i + 1 for i in test_self.elements if i % 2 == 0][: test_self.k]
        )
        test_self.assertTrue(result == expected_results)

    def test_sort_cmp_action_fold(test_self):
        class T(DoAll):
            def sort_cmp(self, a: Any, b: Any) -> int:
                return -1 if a < b else 1 if a > b else 0

            def action(self, chosen):
                return chosen + 1

            def fold(self, results: List) -> Any:
                test_self.assertListEqual(
                    results,
                    sorted([i + 1 for i in test_self.elements])[: test_self.k],
                )
                return sum(results)

        inst = T()
        result = inst(test_self.elements)
        expected_results = sum(
            sorted([i + 1 for i in test_self.elements])[: test_self.k]
        )
        test_self.assertTrue(result == expected_results)

    def test_sort_key_action_fold(test_self):
        class T(DoAll):
            def sort_key(self, chosen: Any):
                return chosen

            def action(self, chosen: Any) -> Any:
                return chosen + 1

            def fold(self, results: List) -> Any:
                test_self.assertListEqual(
                    results,
                    sorted([i + 1 for i in test_self.elements])[: test_self.k],
                )
                return sum(results)

        inst = T()
        result = inst(test_self.elements)
        expected_results = sum(
            sorted([i + 1 for i in test_self.elements])[: test_self.k]
        )
        test_self.assertTrue(result == expected_results)

    def test_filter_sort_cmp_action_fold(test_self):
        class T(DoAll):
            def filter(self, chosen: Any) -> bool:
                return chosen % 2 == 0

            def sort_cmp(self, a: Any, b: Any) -> int:
                return -1 if a < b else 1 if a > b else 0

            def action(self, chosen: Any) -> Any:
                return chosen + 1

            def fold(self, results: List) -> Any:
                test_self.assertListEqual(
                    results,
                    sorted([i + 1 for i in test_self.elements if i % 2 == 0])[
                        : test_self.k
                    ],
                )
                return sum(results)

        inst = T()
        result = inst(test_self.elements)
        expected_results = sum(
            sorted([i + 1 for i in test_self.elements if i % 2 == 0])[: test_self.k]
        )
        test_self.assertTrue(result == expected_results)

    def test_filter_sort_key_action_fold(test_self):
        class T(DoAll):
            def filter(self, chosen: Any) -> bool:
                return chosen % 2 == 0

            def sort_key(self, chosen: Any):
                return chosen

            def action(self, chosen: Any) -> Any:
                return chosen + 1

            def fold(self, results: List) -> Any:
                test_self.assertListEqual(
                    results,
                    sorted([i + 1 for i in test_self.elements if i % 2 == 0])[
                        : test_self.k
                    ],
                )
                return sum(results)

        inst = T()
        result = inst(test_self.elements)
        expected_results = sum(
            sorted([i + 1 for i in test_self.elements if i % 2 == 0])[: test_self.k]
        )
        test_self.assertTrue(result == expected_results)


if __name__ == "__main__":
    unittest.main()



import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
import unittest

from metacontroller import DoOne


class TestDoOneSmoke(unittest.TestCase):
    def test_basic(self):
        class BasicDoOne(DoOne):
            def __init__(self):
                self.pre_controller_passed = False
                self.filter_passed = False
                self.sort_key_passed = False
                self.action_passed = False
                self.post_controller_passed = False

            def pre_controller(self, arg1) -> None:
                self.pre_controller_passed = arg1

            def filter(self, chosen) -> bool:
                self.filter_passed = True
                return True

            def sort_key(self, chosen):
                self.sort_key_passed = True
                return chosen

            def action(self, chosen, arg1: bool):
                self.action_passed = arg1
                return chosen

            def post_controller(self, arg1) -> None:
                self.post_controller_passed = arg1

        inst = BasicDoOne()
        elements = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 0]
        result = inst(elements, True)
        self.assertTrue(result == 0)
        self.assertTrue(inst.pre_controller_passed)
        self.assertTrue(inst.filter_passed)
        self.assertTrue(inst.sort_key_passed)
        self.assertTrue(inst.action_passed)
        self.assertTrue(inst.post_controller_passed)


if __name__ == "__main__":
    unittest.main()



import ast
import inspect
import warnings
from textwrap import dedent
from typing import Any, Callable, List, Tuple


class MethodInspector:
    def __init__(self, fn: Callable) -> None:
        static_override = False
        if not callable(fn):
            abort = True
            try:
                if callable(fn.__func__):
                    abort = False
                    fn = fn.__func__
                    static_override = True
            except BaseException:
                pass

            if abort:
                try:
                    name = fn.__name__
                except AttributeError:
                    name = fn.__func__.__name__
                except BaseException:
                    name = "UNKNOWN"
                raise ValueError(
                    f'MethodInspector expected a callable object, but "{name}" is not.'
                )

        self.fn = fn
        self.__is_staticmethod = isinstance(fn, staticmethod) or static_override
        self.__is_lambda = inspect.isfunction(fn) and fn.__name__ == "<lambda>"

        if hasattr(fn, "__wrapped__"):
            self.spec = inspect.getfullargspec(fn.__wrapped__)
            self._signature_dict = self.signature_to_dict(fn.__wrapped__)
        else:
            self.spec = inspect.getfullargspec(fn)
            self._signature_dict = self.signature_to_dict(fn)

        # set the placeholder values for the return options
        self.__has_explicit_void_return = None
        self.__has_explicit_value_return = None
        self.__has_value_yield = None
        self.__has_value_yield_from = None
        self.__error = None
        self.__decompiled_module = None
        self.__source = None

    ###
    # Read Only Properties
    #

    @property
    def is_staticmethod(self) -> bool:
        return self.__is_staticmethod

    @property
    def is_lambda(self) -> bool:
        return self.__is_lambda

    @property
    def has_explicit_void_return(self) -> bool:
        if self.__has_explicit_void_return is None:
            self._parse_return_options()
        return self.__has_explicit_void_return

    @property
    def has_explicit_value_return(self) -> bool:
        if self.__has_explicit_value_return is None:
            self._parse_return_options()
        if self.is_lambda:
            return True  # lambdas always return a value
        return self.__has_explicit_value_return

    @property
    def has_value_yield(self) -> bool:
        if self.__has_value_yield is None:
            self._parse_return_options()
        return self.__has_value_yield

    @property
    def has_value_yield_from(self) -> bool:
        if self.__has_value_yield_from is None:
            self._parse_return_options()
        return self.__has_value_yield_from

    @property
    def returns_a_value(self) -> bool:
        return (
            self.has_explicit_value_return
            or self.has_value_yield
            or self.has_value_yield_from
        )

    @property
    def has_parse_error(self) -> bool:
        if self.__error is None:
            self._parse_return_options()
        return self.__error

    @property
    def name(self) -> str:
        return self.fn.__name__

    @property
    def posonlyargs(self) -> list:
        """
        Only the position only arguments in the args list.
        """
        return self._signature_dict["posonlyargs"]

    @property
    def args(self) -> list:
        """
        Args are made up of position only and keyword arguments.
        """
        return self._signature_dict["args"] or list()

    @property
    def call_args(self) -> list:
        """
        Call args are made up of arguments excluding the class arg if
        this is an instance method.

        Returns:
            list: _description_
        """
        if self.is_staticmethod:
            return self.args
        return self.args[1:]

    @property
    def num_parameters(self) -> int:
        counter = 0
        counter += len(self.args)
        counter += len(self.kwonlyargs)
        counter += 1 if self.varargs is not None else 0
        counter += 1 if self.varkw is not None else 0
        return counter

    @property
    def num_call_parameters(self) -> int:
        counter = self.num_parameters
        if not self.is_staticmethod:
            counter = max(counter - 1, 0)  # remove the class argument
        return counter

    @property
    def varargs(self) -> str:
        return self._signature_dict["varargs"] or None

    @property
    def varkw(self) -> str:
        return self._signature_dict["varkw"] or None

    @property
    def defaults(self) -> list:
        return self._signature_dict["defaults"] or list()

    @property
    def kwonlyargs(self) -> list:
        return self._signature_dict["kwonlyargs"] or list()

    @property
    def kwonlydefaults(self) -> list:
        return self._signature_dict["kwonlydefaults"] or list()

    @property
    def annotations(self) -> list:
        return self._signature_dict["annotations"] or list()

    @property
    def has_arg_unpack(self):
        return self.varargs is not None

    @property
    def has_kwarg_unpack(self):
        return self.varkw is not None

    @property
    def full_call_arg_spec(self) -> inspect.FullArgSpec:
        if self.__is_staticmethod:
            return self.spec

        # remove the self argument
        return inspect.FullArgSpec(
            args=self.spec.args[1:],
            varargs=self.spec.varargs,
            varkw=self.spec.varkw,
            defaults=self.spec.defaults,
            kwonlyargs=self.spec.kwonlyargs,
            kwonlydefaults=self.spec.kwonlydefaults,
            annotations=self.spec.annotations,
        )

    @property
    def body_ast(self) -> List[ast.AST]:
        """
        Returns the body ast of the method. NOTE: This excludes the signature.

        Returns:
            List[ast.AST]: List of AST nodes representing the body of the method.
        """
        if self.has_parse_error:
            return None
        elif self.is_lambda:
            raise NotImplementedError(
                "Inspecting lambda functions is not supported yet."
            )
            # TODO: REPLACE THIS. I should handle lambdas entirely separetely, and boil them down to this class's properties
            return [self.__decompiled_module.body[0].value.body]
        # get the module body, then the body of the function
        return self.__decompiled_module.body[0].body

    def get_defaulted_args(self) -> List[Tuple[str, Any]]:
        """
        Returns a list of tuples of (str,Any) being the argument name and its default.

        Returns:
            List[Tuple[str, Any]]: defaulted argument names and values
        """
        keywords = self.args[len(self.args) - len(self.defaults) :]
        if len(keywords) == 0:
            return list()

        return list(zip(keywords, self.defaults))

    def get_keyword_only_args(self) -> List[Tuple[str, Any]]:
        """
        Returns a list of tuples of (str,Any) being the keyword and its value.

        Returns:
            List[Tuple]: keyword only argument names and values
        """
        if len(self.kwonlyargs) == 0:
            return list()

        return list(self.kwonlydefaults.items())

    def get_non_defaulted_args(self) -> List[str]:
        """
        Returns a the list of arguments that are either position only, or non
        defaulted.

        Returns:
            List[str]: Position only or non-defaulted arguments.
        """
        return self.args[: -len(self.get_defaulted_args()) or None]

    def _parse_return_options(self) -> None:
        """Inspection method to parse this instances' callable and determine the
        different ways it can exit:

        Explicit Void returns are when a return statement with no value is provided in the top
        level scope of the provided method.

        Explicit Value returns are when a return statement with a right hand value
        (including None) is provided in the top level scope of the method.

        Value Yield and Value Yield From are equivalent to the checks above, but for
        the yield and yield from keywords.
        """
        # pre-set the values
        self.__has_explicit_void_return = False
        self.__has_explicit_value_return = False
        self.__has_value_yield = False
        self.__has_value_yield_from = False
        self.__error = False

        try:
            self.__source = inspect.getsource(self.fn)
            self.__decompiled_module = ast.parse(dedent(self.__source))

            class InnerReturnVisitor(ast.NodeVisitor):
                def __init__(self):
                    self.has_explicit_void_return = False
                    self.has_explicit_value_return = False
                    self.has_value_yield = False
                    self.has_value_yield_from = False
                    self.__func_hit = False
                    self.__parent_map = {}

                def visit(self, node):
                    if isinstance(node, list):
                        for item in node:
                            if isinstance(item, ast.AST):
                                self.__parent_map[item] = node
                                super().visit(item)
                    elif isinstance(node, ast.AST):
                        for child in ast.iter_child_nodes(node):
                            self.__parent_map[child] = node
                        super().visit(node)

                def visit_Return(self, node):
                    if node.value is not None:
                        self.has_explicit_value_return = True
                    else:
                        self.has_explicit_void_return = True

                def visit_Yield(self, node: ast.Yield):
                    if node.value is not None:
                        self.has_value_yield = True

                def visit_YieldFrom(self, node: ast.YieldFrom):
                    if node.value is not None:
                        self.has_value_yield_from = True

                def visit_FunctionDef(self, node):
                    if self.__func_hit == False:
                        self.__func_hit = True
                        self.generic_visit(node)
                    else:
                        next_node = self.get_next_node(node)
                        if next_node is not None:
                            self.visit(next_node)

                def visit_AsyncFunctionDef(self, node):
                    next_node = self.get_next_node(node)
                    if next_node is not None:
                        self.visit(next_node)

                def visit_Lambda(self, node):
                    next_node = self.get_next_node(node)
                    if next_node is not None:
                        self.visit(next_node)

                def visit_ClassDef(self, node: ast.ClassDef):
                    next_node = self.get_next_node(node)
                    if next_node is not None:
                        self.visit(next_node)

                def get_next_node(self, node):
                    parent = self.__parent_map[node]
                    if hasattr(parent, "body"):
                        next_sibling = (
                            parent.body[parent.body.index(node) + 1]
                            if parent.body.index(node) + 1 < len(parent.body)
                            else None
                        )
                    else:
                        next_sibling = None
                    return next_sibling

            visitor = InnerReturnVisitor()
            visitor.visit(
                self.__decompiled_module.body[0]
            )  # Only visit the top-level function

            self.__has_explicit_value_return = visitor.has_explicit_value_return
            self.__has_explicit_void_return = visitor.has_explicit_void_return
            self.__has_value_yield = visitor.has_value_yield
            self.__has_value_yield_from = visitor.has_value_yield_from
            self.__error = False

        except BaseException as err:
            try:
                name = self.fn.__name__
            except BaseException:
                name = "UNKNOWN"
            warnings.warn(f'Unable to parse callable "{name}". Error message: {err}')
            self.__error = True

    @staticmethod
    def signature_to_dict(fn: Callable) -> dict:
        """
        returns a dict with the following keys:
        'posonlyargs', 'args', 'varargs', 'varkw', 'defaults', 'kwonlyargs', 'kwonlydefaults', 'annotations'

        NOTE: if "posonlyargs" exists, they will also be found in "args".

        Args:
            fn (Callable): callable object

        Returns:
            dict: dictionary with the components of the call signature
        """
        result = {}
        result["posonlyargs"] = [
            name
            for name, param in inspect.signature(fn).parameters.items()
            if param.kind == inspect.Parameter.POSITIONAL_ONLY
        ]
        result.update(inspect.getfullargspec(fn)._asdict())
        return result


if __name__ == "__main__":

    def test():
        class TestClass:
            def test2(self):
                return 0

            __call__ = lambda: print("called")

        yield (lambda: TestClass)()

    # import timeit
    # print(timeit.timeit(lambda: MethodInspector(test).has_parse_error, number=1000))

    lambda_test = lambda: print("test")
    r = MethodInspector(lambda_test)
    syntax = r.body_ast
    result = MethodInspector(test)
    print(result.has_explicit_value_return)
    print(result.has_explicit_void_return)
    print(result.has_value_yield)
    print(result.has_value_yield_from)
    print(result.has_parse_error)



import ast
from abc import ABC, abstractmethod
from textwrap import dedent
from typing import Any, Callable, Dict, List, Tuple, Union

from metacontroller.internal.exceptions import (
    ArgumentError,
    InvalidControllerMethodError,
)
from metacontroller.internal.method_inspector import MethodInspector
from metacontroller.internal.namespace import (
    ACTION_METHOD_NAME,
    CLASS_ARG_NAME,
    FILTER_METHOD_NAME,
    FOLD_METHOD_NAME,
    GENERATED_CALL_METHOD_NAME,
    K_ARG_NAME,
    PARTITION_ARG_NAME,
    POST_CONTROLLER_METHOD_NAME,
    PRE_CONTROLLER_METHOD_NAME,
    SORT_CMP_METHOD_NAME,
    SORT_KEY_METHOD_NAME,
)


class BaseControllerImplementation(ABC):
    def __init__(
        self,
        cls,
        name,
        bases,
        attrs,
        stack_frame,
        pre_controller_enabled: bool = True,
        filter_enabled: bool = True,
        sort_key_enabled: bool = True,
        sort_cmp_enabled: bool = True,
        action_enabled: bool = True,
        fold_enabled: bool = True,
        post_controller_enabled: bool = True,
    ) -> None:
        super().__init__()
        self.cls = cls
        self.name = name
        self.bases = bases
        self.attrs = attrs
        self.stack_frame = stack_frame

        self.__pre_controller = (
            MethodInspector(self.attrs[PRE_CONTROLLER_METHOD_NAME])
            if PRE_CONTROLLER_METHOD_NAME in self.attrs and pre_controller_enabled
            else None
        )

        self.__filter = (
            MethodInspector(self.attrs[FILTER_METHOD_NAME])
            if FILTER_METHOD_NAME in self.attrs and filter_enabled
            else None
        )

        self.__sort_key = (
            MethodInspector(self.attrs[SORT_KEY_METHOD_NAME])
            if SORT_KEY_METHOD_NAME in self.attrs and sort_key_enabled
            else None
        )

        self.__sort_cmp = (
            MethodInspector(self.attrs[SORT_CMP_METHOD_NAME])
            if SORT_CMP_METHOD_NAME in self.attrs and sort_cmp_enabled
            else None
        )

        self.__action = (
            MethodInspector(self.attrs[ACTION_METHOD_NAME])
            if ACTION_METHOD_NAME in self.attrs and action_enabled
            else None
        )

        self.__fold = (
            MethodInspector(self.attrs[FOLD_METHOD_NAME])
            if FOLD_METHOD_NAME in self.attrs and fold_enabled
            else None
        )

        self.__post_controller = (
            MethodInspector(self.attrs[POST_CONTROLLER_METHOD_NAME])
            if POST_CONTROLLER_METHOD_NAME in self.attrs and post_controller_enabled
            else None
        )

    ####
    # Methods to be defined by sub-classes

    @abstractmethod
    def validate(self) -> None:
        """
        Validation method to ensure that class attributes and method signatures/return values
        are all valid before any compilation is done. This method should raise specific
        exceptions if a controller is deemed invalid.
        """
        # assert there is some sort of controlled method in the controller
        if (
            not self.has_pre_controller
            and not self.has_filter
            and not self.has_sort_key
            and not self.has_sort_cmp
            and not self.has_action
            and not self.has_fold
            and not self.has_post_controller
        ):
            raise InvalidControllerMethodError(
                f'"{self.name}" must have at least one controlled method to be valid.'
            )

    @abstractmethod
    def generate_call_method(self) -> Callable[..., Any]:
        """
        Delegate to each controller implementation to create its own call method.
        This must return a callable method that will be bound to the classes __call__
        dunder method. Each instance will be callable with the output of this method.

        Returns:
            Callable[..., Any]: __call__() method for the controller instances.
        """
        ...

    ####
    # Read only properties

    @property
    def has_pre_controller(self) -> bool:
        return self.__pre_controller is not None

    @property
    def pre_controller(self) -> Union[MethodInspector, None]:
        return self.__pre_controller

    @property
    def has_filter(self) -> bool:
        return self.__filter is not None

    @property
    def filter(self) -> Union[MethodInspector, None]:
        return self.__filter

    @property
    def has_sort_key(self) -> bool:
        return self.__sort_key is not None

    @property
    def sort_key(self) -> Union[MethodInspector, None]:
        return self.__sort_key

    @property
    def has_sort_cmp(self) -> bool:
        return self.__sort_cmp is not None

    @property
    def sort_cmp(self) -> Union[MethodInspector, None]:
        return self.__sort_cmp

    @property
    def has_action(self) -> bool:
        return self.__action is not None

    @property
    def action(self) -> Union[MethodInspector, None]:
        return self.__action

    @property
    def has_fold(self) -> bool:
        return self.__fold is not None

    @property
    def fold(self) -> Union[MethodInspector, None]:
        return self.__fold

    @property
    def has_post_controller(self) -> bool:
        return self.__post_controller is not None

    @property
    def post_controller(self) -> Union[MethodInspector, None]:
        return self.__post_controller

    ####
    # Common helpers

    def compile_call_method(
        self, module: ast.Module, additional_globals: dict = None
    ) -> Callable[..., Any]:
        """
        Compiles the call method from within the passed in module for this controller.

        Args:
            module (ast.Module): module which contains the call function.
            additional_globals (dict, optional): additional global values to include in compilation. Defaults to None.

        Returns:
            Callable[..., Any]: generated call method for this controller.
        """
        _globals = self.stack_frame.f_globals
        _globals.update(self.stack_frame.f_locals)
        if additional_globals is not None:
            _globals.update(additional_globals)

        _locals = {}
        eval(
            compile(module, filename="<ast>", mode="exec"),
            _globals,
            _locals,
        )
        return _locals[GENERATED_CALL_METHOD_NAME]

    def get_call_args(
        self,
        use_class_arg: bool = True,
        use_k_arg: bool = False,
        use_partition_arg: bool = True,
        required_pre_controller_args: int = 0,
        required_filter_args: int = 1,
        required_sort_key_args: int = 1,
        required_sort_cmp_args: int = 2,
        required_action_args: int = 1,
        requried_fold_args: int = 1,
        required_post_controller_args: int = 0,
    ) -> Tuple[ast.arguments, dict]:
        """
        The goal of this function is to determine the set of parameters that make up the call method
        of this controller. This is done by inspecting the paramters of the controlled methods
        that are a part of this controller.

        Position only and non-defaulted keyword arguments are required to share the same name.

        Keyword only and default arguments are allowed to shared the same name, but they must also
        share the same default argument. Duplicated keywords with different default values
        are invalid, and an error will be thrown.

        Variable args are supported, but any controlled method that uses them will be required to use
        the same vararg name.

        Kwargs are supported, but any controlled method that uses them will be required to use
        the same kwarg name.

        Raises:
            AttributeError: for duplicate keyword/default IDs without the same default value.

        Returns:
            Tuple[ast.arguments, dict]: A tuple containing the Arguments for the generated call method, as
            well as a dictionary of the saved argument defaults that should be passed to the compile method.
        """
        # arguments to be used when constructing the resulting ast.arguments
        posonlyargs = []
        args = []
        kwonlyargs = []
        kw_defaults = []
        defaults = []
        var_arg = None
        kwarg = None
        saved_defaults = {}

        if use_class_arg:
            # add the class argument
            posonlyargs.append(
                ast.arg(arg=CLASS_ARG_NAME, annotation=None, type_comment=None)
            )

        if use_k_arg:
            # add the dynamic_max_chosen K argument
            posonlyargs.append(
                ast.arg(arg=K_ARG_NAME, annotation=None, type_comment=None)
            )

        if use_partition_arg:
            # add the partition argument
            posonlyargs.append(
                ast.arg(arg=PARTITION_ARG_NAME, annotation=None, type_comment=None)
            )

        # join the positional and non-defaulted arguments from the controlled methods
        pre_controller_args = []
        if self.has_pre_controller:
            arg_start_index = 0 if self.pre_controller.is_staticmethod else 1
            arg_start_index += required_pre_controller_args
            pre_controller_args = self.pre_controller.get_non_defaulted_args()[
                arg_start_index:
            ]

        filter_args = []
        if self.has_filter:
            arg_start_index = 0 if self.filter.is_staticmethod else 1
            arg_start_index += required_filter_args
            filter_args = self.filter.get_non_defaulted_args()[arg_start_index:]

        sort_key_args = []
        if self.has_sort_key:
            arg_start_index = 0 if self.sort_key.is_staticmethod else 1
            arg_start_index += required_sort_key_args
            sort_key_args = self.sort_key.get_non_defaulted_args()[arg_start_index:]

        sort_cmp_args = []
        if self.has_sort_cmp:
            arg_start_index = 0 if self.sort_cmp.is_staticmethod else 1
            arg_start_index += required_sort_cmp_args
            sort_cmp_args = self.sort_cmp.get_non_defaulted_args()[arg_start_index:]

        action_args = []
        if self.has_action:
            arg_start_index = 0 if self.action.is_staticmethod else 1
            arg_start_index += required_action_args
            action_args = self.action.get_non_defaulted_args()[arg_start_index:]

        fold_args = []
        if self.has_fold:
            arg_start_index = 0 if self.fold.is_staticmethod else 1
            arg_start_index += requried_fold_args
            fold_args = self.fold.get_non_defaulted_args()[arg_start_index:]

        post_controller_args = []
        if self.has_post_controller:
            arg_start_index = 0 if self.post_controller.is_staticmethod else 1
            arg_start_index += required_post_controller_args
            post_controller_args = self.post_controller.get_non_defaulted_args()[
                arg_start_index:
            ]

        max_args = max(
            len(pre_controller_args),
            len(filter_args),
            len(sort_key_args),
            len(sort_cmp_args),
            len(action_args),
            len(fold_args),
            len(post_controller_args),
        )

        # check each non-defaulted argument in each index and ensure its the same argument name
        shared_msg = "Shared positional arguments must have the same name across all controlled methods that use it."
        for index in range(0, max_args, 1):
            current_arg: Union[str, None] = None
            if len(pre_controller_args) > index:
                current_arg = pre_controller_args[index]

            if len(filter_args) > index:
                if current_arg is None:
                    current_arg = filter_args[index]
                elif current_arg != filter_args[index]:
                    msg = f'{FILTER_METHOD_NAME} argument {index} "{filter_args[index]}" is positionally shared with "{current_arg}"; choose one name for this argument.'
                    msg += shared_msg
                    raise ArgumentError(msg)

            if len(sort_key_args) > index:
                if current_arg is None:
                    current_arg = sort_key_args[index]
                elif current_arg != sort_key_args[index]:
                    msg = f'{SORT_KEY_METHOD_NAME} argument {index} "{sort_key_args[index]}" is positionally shared with "{current_arg}"; choose one name for this argument. '
                    msg += shared_msg
                    raise ArgumentError(msg)

            if len(sort_cmp_args) > index:
                if current_arg is None:
                    current_arg = sort_cmp_args[index]
                elif current_arg != sort_cmp_args[index]:
                    msg = f'{SORT_CMP_METHOD_NAME} argument {index} "{sort_cmp_args[index]}" is positionally shared with "{current_arg}"; choose one name for this argument. '
                    msg += shared_msg
                    raise ArgumentError(msg)

            if len(action_args) > index:
                if current_arg is None:
                    current_arg = action_args[index]
                elif current_arg != action_args[index]:
                    msg = f'{ACTION_METHOD_NAME} argument {index} "{action_args[index]}" is positionally shared with "{current_arg}"; choose one name for this argument. '
                    msg += shared_msg
                    raise ArgumentError(msg)

            if len(fold_args) > index:
                if current_arg is None:
                    current_arg = fold_args[index]
                elif current_arg != fold_args[index]:
                    msg = f'{FOLD_METHOD_NAME} argument {index} "{fold_args[index]}" is positionally shared with "{current_arg}"; choose one name for this argument. '
                    msg += shared_msg
                    raise ArgumentError(msg)

            if len(post_controller_args) > index:
                if current_arg is None:
                    current_arg = post_controller_args[index]
                elif current_arg != post_controller_args[index]:
                    msg = f'{POST_CONTROLLER_METHOD_NAME} argument {index} "{post_controller_args[index]}" is positionally shared with "{current_arg}"; choose one name for this argument. '
                    msg += shared_msg
                    raise ArgumentError(msg)

            if current_arg is not None:
                args.append(ast.arg(arg=current_arg, annotation=None))

        # defined as an inner function to pass args and saved_global_kwargs into the namespace
        def _should_include_arg(
            keyword: str,
            default: Any,
            existing_args: List[str],
            saved_global_kwargs: Dict[str, Any] = saved_defaults,
        ) -> bool:
            """
            Checks if a keyword argument already exists with the same default value. If it exists,
            return False since the keyword argument is already taken care of. If it does not exist,
            return True since it should be added. If the keyword argument exists and has a different
            (decided by == (__eq__) operator) default value, raise an exception because they must be
            the same default.

            Args:
                keyword (str): keyword argument in question
                default (Any): default value for the keyword in question
                existing_args (List[str]): arguments that have already been added.
                saved_global_kwargs (Dict[str, Any], optional): saved defaults. Defaults to self.saved_global_kwargs.

            Raises:
                AttributeError: Duplicate keyword arguments must have equivalent default values.

            Returns:
                bool: Should include in args list
            """
            if keyword in existing_args:
                _ctrl_keyword_name = f"{keyword}"
                if _ctrl_keyword_name in saved_global_kwargs:
                    val = saved_global_kwargs[_ctrl_keyword_name]
                if default == val:
                    return False
                else:
                    raise AttributeError(
                        dedent(
                            f'Duplicate keyword argument "{keyword}" with different default values.\
                            Shared keyword arguments must have the same default value.\
                                Equality is checked with the __eq__ operator.'
                        )
                    )
            return True

        def add_non_conflicting_parameters(
            keyword_values: List[Tuple[str, Any]], args: list, defaults: list
        ) -> None:
            for keyword, value in keyword_values:
                if _should_include_arg(keyword, value, [arg.arg for arg in args]):
                    args.append(ast.arg(arg=keyword, annotation=None))
                    global_keyword_name = f"{keyword}"
                    saved_defaults[global_keyword_name] = value
                    defaults.append(ast.Name(id=global_keyword_name, ctx=ast.Load()))

        # get the defaulted arguments and keyword only arguments
        if self.has_pre_controller:
            add_non_conflicting_parameters(
                self.pre_controller.get_defaulted_args(), args, defaults
            )
            add_non_conflicting_parameters(
                self.pre_controller.get_keyword_only_args(), kwonlyargs, kw_defaults
            )

        if self.has_filter:
            add_non_conflicting_parameters(
                self.filter.get_defaulted_args(), args, defaults
            )
            add_non_conflicting_parameters(
                self.filter.get_keyword_only_args(), kwonlyargs, kw_defaults
            )

        if self.has_sort_key:
            add_non_conflicting_parameters(
                self.sort_key.get_defaulted_args(), args, defaults
            )
            add_non_conflicting_parameters(
                self.sort_key.get_keyword_only_args(),
                kwonlyargs,
                kw_defaults,
            )

        if self.has_sort_cmp:
            add_non_conflicting_parameters(
                self.sort_cmp.get_defaulted_args(), args, defaults
            )
            add_non_conflicting_parameters(
                self.sort_cmp.get_keyword_only_args(),
                kwonlyargs,
                kw_defaults,
            )

        if self.has_action:
            add_non_conflicting_parameters(
                self.action.get_defaulted_args(), args, defaults
            )
            add_non_conflicting_parameters(
                self.action.get_keyword_only_args(), kwonlyargs, kw_defaults
            )

        if self.has_fold:
            add_non_conflicting_parameters(
                self.fold.get_defaulted_args(), args, defaults
            )
            add_non_conflicting_parameters(
                self.fold.get_keyword_only_args(), kwonlyargs, kw_defaults
            )

        if self.has_post_controller:
            add_non_conflicting_parameters(
                self.post_controller.get_defaulted_args(), args, defaults
            )
            add_non_conflicting_parameters(
                self.post_controller.get_keyword_only_args(), kwonlyargs, kw_defaults
            )

        # check for arg unpacks
        arg_unpack_name = None
        if self.has_pre_controller:
            arg_unpack_name = self.pre_controller.varargs
        if self.has_filter:
            if arg_unpack_name is None:
                arg_unpack_name = self.filter.varargs
            elif self.filter.has_arg_unpack and self.filter.varargs != arg_unpack_name:
                raise ArgumentError(
                    dedent(
                        f'{FILTER_METHOD_NAME} controlled action uses "{self.filter.varargs}" as the argument unpack variable name, \
                    but it was previously defined as "{arg_unpack_name}". \
                    The argument unpack variable must be the same name across all controlled methods that use it.'
                    )
                )
        if self.has_sort_key:
            if arg_unpack_name is None:
                arg_unpack_name = self.sort_key.varargs
            elif (
                self.sort_key.has_arg_unpack
                and self.sort_key.varargs != arg_unpack_name
            ):
                raise ArgumentError(
                    dedent(
                        f'{SORT_KEY_METHOD_NAME} controlled action uses "{self.sort_key.varargs}" as the argument unpack variable name, \
                    but it was previously defined as "{arg_unpack_name}". \
                    The argument unpack variable must be the same name across all controlled methods that use it.'
                    )
                )
        if self.has_sort_cmp:
            if arg_unpack_name is None:
                arg_unpack_name = self.sort_cmp.varargs
            elif (
                self.sort_cmp.has_arg_unpack
                and self.sort_cmp.varargs != arg_unpack_name
            ):
                raise ArgumentError(
                    dedent(
                        f'{SORT_CMP_METHOD_NAME} controlled action uses "{self.sort_cmp.varargs}" as the argument unpack variable name, \
                    but it was previously defined as "{arg_unpack_name}". \
                    The argument unpack variable must be the same name across all controlled methods that use it.'
                    )
                )
        if self.has_action:
            if arg_unpack_name is None:
                arg_unpack_name = self.action.varargs
            elif self.action.has_arg_unpack and self.action.varargs != arg_unpack_name:
                raise ArgumentError(
                    dedent(
                        f'{ACTION_METHOD_NAME} controlled action uses "{self.action.varargs}" as the argument unpack variable name, \
                    but it was previously defined as "{arg_unpack_name}". \
                    The argument unpack variable must be the same name across all controlled methods that use it.'
                    )
                )
        if self.has_fold:
            if arg_unpack_name is None:
                arg_unpack_name = self.fold.varargs
            elif self.fold.has_arg_unpack and self.fold.varargs != arg_unpack_name:
                raise ArgumentError(
                    dedent(
                        f'{FOLD_METHOD_NAME} controlled action uses "{self.fold.varargs}" as the argument unpack variable name, \
                    but it was previously defined as "{arg_unpack_name}". \
                    The argument unpack variable must be the same name across all controlled methods that use it.'
                    )
                )
        if self.has_post_controller:
            if arg_unpack_name is None:
                arg_unpack_name = self.post_controller.varargs
            elif (
                self.post_controller.has_arg_unpack
                and self.post_controller.varargs != arg_unpack_name
            ):
                raise ArgumentError(
                    dedent(
                        f'{POST_CONTROLLER_METHOD_NAME} controlled action uses "{self.post_controller.varargs}" as the argument unpack variable name, \
                    but it was previously defined as "{arg_unpack_name}". \
                    The argument unpack variable must be the same name across all controlled methods that use it.'
                    )
                )

        if arg_unpack_name is not None:
            var_arg = ast.arg(arg=arg_unpack_name, annotation=None)

        # check for kwarg unpacks
        kwarg_name = None
        if self.has_pre_controller:
            kwarg_name = self.pre_controller.varkw
        if self.has_filter:
            if kwarg_name is None:
                kwarg_name = self.filter.varkw
            elif self.filter.has_kwarg_unpack and self.filter.varkw != kwarg_name:
                raise ArgumentError(
                    dedent(
                        f'{FILTER_METHOD_NAME} controlled action uses "{self.filter.varkw}" as the keyword argument unpack variable name, \
                    but it was previously defined as "{kwarg_name}". \
                    The keyword argument unpack variable must be the same name across all controlled methods that use it.'
                    )
                )
        if self.has_sort_key:
            if kwarg_name is None:
                kwarg_name = self.sort_key.varkw
            elif self.sort_key.has_kwarg_unpack and self.sort_key.varkw != kwarg_name:
                raise ArgumentError(
                    dedent(
                        f'{SORT_KEY_METHOD_NAME} controlled action uses "{self.sort_key.varkw}" as the keyword argument unpack variable name, \
                    but it was previously defined as "{kwarg_name}". \
                    The keyword argument unpack variable must be the same name across all controlled methods that use it.'
                    )
                )
        if self.has_sort_cmp:
            if kwarg_name is None:
                kwarg_name = self.sort_cmp.varkw
            elif self.sort_cmp.has_kwarg_unpack and self.sort_cmp.varkw != kwarg_name:
                raise ArgumentError(
                    dedent(
                        f'{SORT_CMP_METHOD_NAME} controlled action uses "{self.sort_cmp.varkw}" as the keyword argument unpack variable name, \
                    but it was previously defined as "{kwarg_name}". \
                    The keyword argument unpack variable must be the same name across all controlled methods that use it.'
                    )
                )
        if self.has_action:
            if kwarg_name is None:
                kwarg_name = self.action.varkw
            elif self.action.has_kwarg_unpack and self.action.varkw != kwarg_name:
                raise ArgumentError(
                    dedent(
                        f'{ACTION_METHOD_NAME} controlled action uses "{self.action.varkw}" as the keyword argument unpack variable name, \
                    but it was previously defined as "{kwarg_name}". \
                    The keyword argument unpack variable must be the same name across all controlled methods that use it.'
                    )
                )
        if self.has_fold:
            if kwarg_name is None:
                kwarg_name = self.fold.varkw
            elif self.fold.has_kwarg_unpack and self.fold.varkw != kwarg_name:
                raise ArgumentError(
                    dedent(
                        f'{FOLD_METHOD_NAME} controlled action uses "{self.fold.varkw}" as the keyword argument unpack variable name, \
                    but it was previously defined as "{kwarg_name}". \
                    The keyword argument unpack variable must be the same name across all controlled methods that use it.'
                    )
                )
        if self.has_post_controller:
            if kwarg_name is None:
                kwarg_name = self.post_controller.varkw
            elif (
                self.post_controller.has_kwarg_unpack
                and self.post_controller.varkw != kwarg_name
            ):
                raise ArgumentError(
                    dedent(
                        f'{POST_CONTROLLER_METHOD_NAME} controlled action uses "{self.post_controller.varkw}" as the keyword argument unpack variable name, \
                    but it was previously defined as "{kwarg_name}". \
                    The keyword argument unpack variable must be the same name across all controlled methods that use it.'
                    )
                )

        if kwarg_name is not None:
            kwarg = ast.arg(arg=kwarg_name, annotation=None)

        # create the signatures arguments
        func_args = ast.arguments(
            posonlyargs=posonlyargs,
            args=args,
            vararg=var_arg,
            kwonlyargs=kwonlyargs,
            kw_defaults=kw_defaults,
            kwarg=kwarg,
            defaults=defaults,
        )
        return func_args, saved_defaults



import random
from typing import Any

random.seed(0)
import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
import unittest

from metacontroller import DoOne


class ArgCheck:
    def __init__(
        self,
        filter_passed=False,
        sort_cmp_passed=False,
        sort_key_passed=False,
        action_passed=False,
        fold_passed=False,
    ) -> None:
        self.filter_passed = filter_passed
        self.sort_cmp_passed = sort_cmp_passed
        self.sort_key_passed = sort_key_passed
        self.action_passed = action_passed
        self.fold_passed = fold_passed


class TestDoOneMethodCombinationsWithArg(unittest.TestCase):
    def setUp(test_self):
        test_self.elements = [random.randint(0, 1000) for _ in range(10)]

    def test_filter_sort_cmp(test_self):
        class T(DoOne):
            def filter(self, chosen, arg) -> bool:
                arg.filter_passed = True
                return chosen % 2 == 0

            def sort_cmp(self, a, b, arg):
                arg.sort_cmp_passed = True
                return -1 if a < b else 1 if a > b else 0

        arg = ArgCheck()
        inst = T()
        expected_result = sorted([i for i in test_self.elements if i % 2 == 0])[0]
        result = inst(test_self.elements, arg)
        test_self.assertTrue(
            result == expected_result,
        )
        test_self.assertTrue(arg.filter_passed)
        test_self.assertTrue(arg.sort_cmp_passed)

    def test_filter_sort_key(test_self):
        class T(DoOne):
            def filter(self, chosen, arg) -> bool:
                arg.filter_passed = True
                return chosen % 2 == 0

            def sort_key(self, chosen, arg):
                arg.sort_key_passed = True
                return chosen

        arg = ArgCheck()
        inst = T()
        result = inst(test_self.elements, arg)
        test_self.assertTrue(
            result == sorted(list(filter(lambda x: x % 2 == 0, test_self.elements)))[0]
        )
        test_self.assertTrue(arg.filter_passed)
        test_self.assertTrue(arg.sort_key_passed)

    def test_filter_action(test_self):
        class T(DoOne):
            def filter(self, chosen, arg) -> bool:
                arg.filter_passed = True
                return chosen % 2 == 0

            def action(self, chosen: Any, arg) -> Any:
                arg.action_passed = True
                return chosen + 1

        arg = ArgCheck()
        inst = T()
        result = inst(test_self.elements, arg)
        test_self.assertTrue(
            result == [i + 1 for i in test_self.elements if i % 2 == 0][0]
        )
        test_self.assertTrue(arg.filter_passed)
        test_self.assertTrue(arg.action_passed)

    def test_sort_cmp_action(test_self):
        class T(DoOne):
            def sort_cmp(self, a: Any, b: Any, arg) -> int:
                arg.sort_cmp_passed = True
                return -1 if a < b else 1 if a > b else 0

            def action(self, chosen: Any, arg) -> Any:
                arg.action_passed = True
                return chosen + 1

        arg = ArgCheck()
        inst = T()
        result = inst(test_self.elements, arg)
        expected_result = sorted([i + 1 for i in test_self.elements])[0]
        test_self.assertTrue(result == expected_result)
        test_self.assertTrue(arg.sort_cmp_passed)
        test_self.assertTrue(arg.action_passed)

    def test_sort_key_action(test_self):
        class T(DoOne):
            def sort_key(self, chosen: Any, arg):
                arg.sort_key_passed = True
                return chosen

            def action(self, chosen, arg):
                arg.action_passed = True
                return chosen + 1

        arg = ArgCheck()
        inst = T()
        result = inst(test_self.elements, arg)
        test_self.assertTrue(result == [i + 1 for i in sorted(test_self.elements)][0])
        test_self.assertTrue(arg.sort_key_passed)
        test_self.assertTrue(arg.action_passed)

    def test_filter_sort_cmp_action(test_self):
        class T(DoOne):
            def filter(self, chosen, arg) -> bool:
                arg.filter_passed = True
                return chosen % 2 == 0

            def sort_cmp(self, a: Any, b: Any, arg) -> int:
                arg.sort_cmp_passed = True
                return -1 if a < b else 1 if a > b else 0

            def action(self, chosen, arg):
                arg.action_passed = True
                return chosen + 1

        arg = ArgCheck()
        inst = T()
        result = inst(test_self.elements, arg)
        expected_results = sorted([i + 1 for i in test_self.elements if i % 2 == 0])[0]
        test_self.assertTrue(result == expected_results)
        test_self.assertTrue(arg.filter_passed)
        test_self.assertTrue(arg.sort_cmp_passed)
        test_self.assertTrue(arg.action_passed)

    def test_filter_sort_key_action(test_self):
        class T(DoOne):
            def filter(self, chosen, arg) -> bool:
                arg.filter_passed = True
                return chosen % 2 == 0

            def sort_key(self, chosen: Any, arg):
                arg.sort_key_passed = True
                return chosen

            def action(self, chosen: Any, arg) -> Any:
                arg.action_passed = True
                return chosen + 1

        arg = ArgCheck()
        inst = T()
        result = inst(test_self.elements, arg)
        expected_result = sorted([i + 1 for i in test_self.elements if i % 2 == 0])[0]
        test_self.assertTrue(result == expected_result)
        test_self.assertTrue(arg.filter_passed)
        test_self.assertTrue(arg.sort_key_passed)
        test_self.assertTrue(arg.action_passed)


if __name__ == "__main__":
    unittest.main()



import random
from typing import Any, List

random.seed(0)
import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
import unittest

from metacontroller import Do, DoAll, DoK, DoOne
from metacontroller.internal.exceptions import InvalidControllerMethodError


class ArgWrapper:
    def __init__(self, value):
        self.value = value


class TestPreController(unittest.TestCase):
    def test_do(self):
        class T(Do):
            passed = False

            def pre_controller(self) -> None:
                self.passed = True

        inst = T()
        inst()
        self.assertTrue(inst.passed)

    def test_do_one(self):
        class T(DoOne):
            passed = False

            def pre_controller(self) -> None:
                self.passed = True

        inst = T()
        inst([])
        self.assertTrue(inst.passed)

    def test_do_k(self):
        class T(DoK):
            passed = False

            def pre_controller(self) -> None:
                self.passed = True

        inst = T()
        inst(0, [])
        self.assertTrue(inst.passed)

    def test_do_all(self):
        class T(DoAll):
            passed = False

            def pre_controller(self) -> None:
                self.passed = True

        inst = T()
        inst([])
        self.assertTrue(inst.passed)

    # static method tests
    def test_do_static(self):
        class T(Do):
            @staticmethod
            def pre_controller(arg1) -> None:
                arg1.value = True

        inst = T()
        arg = ArgWrapper(False)
        inst(arg)
        self.assertTrue(arg.value)

    def test_do_one_static(self):
        class T(DoOne):
            @staticmethod
            def pre_controller(arg1) -> None:
                arg1.value = True

        inst = T()
        arg = ArgWrapper(False)
        inst([], arg)
        self.assertTrue(arg.value)

    def test_do_k_static(self):
        class T(DoK):
            @staticmethod
            def pre_controller(arg1) -> None:
                arg1.value = True

        inst = T()
        arg = ArgWrapper(False)
        inst(0, [], arg)
        self.assertTrue(arg.value)

    def test_do_all_static(self):
        class T(DoAll):
            @staticmethod
            def pre_controller(arg1) -> None:
                arg1.value = True

        inst = T()
        arg = ArgWrapper(False)
        inst([], arg)
        self.assertTrue(arg.value)


class TestFilter(unittest.TestCase):
    def test_do(self):
        with self.assertRaises(InvalidControllerMethodError):

            class T(Do):
                def filter(self, chosen) -> bool:
                    return chosen % 2 == 0

    def test_do_one(self):
        class T(DoOne):
            def filter(self, chosen) -> bool:
                return chosen % 2 == 0

        inst = T()
        self.assertTrue(inst(range(10)) == 0)

    def test_do_k(self):
        class T(DoK):
            def filter(self, chosen) -> bool:
                return chosen % 2 == 0

        inst = T()
        self.assertTrue(inst(3, range(10)) == [0, 2, 4])

    def test_do_all(self):
        class T(DoAll):
            def filter(self, chosen) -> bool:
                return chosen % 2 == 0

        inst = T()
        self.assertTrue(inst(range(10)) == [0, 2, 4, 6, 8])

    def test_do_one_static(self):
        class T(DoOne):
            @staticmethod
            def filter(chosen) -> bool:
                return chosen % 2 == 0

        inst = T()
        self.assertTrue(inst(range(10)) == 0)

    def test_do_k_static(self):
        class T(DoK):
            @staticmethod
            def filter(chosen) -> bool:
                return chosen % 2 == 0

        inst = T()
        self.assertTrue(inst(3, range(10)) == [0, 2, 4])

    def test_do_all_static(self):
        class T(DoAll):
            @staticmethod
            def filter(chosen) -> bool:
                return chosen % 2 == 0

        inst = T()
        self.assertTrue(inst(range(10)) == [0, 2, 4, 6, 8])


class TestSortKey(unittest.TestCase):

    def setUp(self):
        self.elements = [random.randint(0, 1000) for _ in range(10)]

    def test_do(self):
        with self.assertRaises(InvalidControllerMethodError):

            class T(Do):
                def sort_key(self, chosen) -> int:
                    return chosen

    def test_do_one(self):
        class T(DoOne):
            def sort_key(self, chosen) -> int:
                return chosen

        inst = T()
        self.assertTrue(inst(self.elements) == min(self.elements))

    def test_do_k(self):
        class T(DoK):
            def sort_key(self, chosen) -> int:
                return chosen

        inst = T()
        self.assertTrue(
            inst(3, self.elements) == sorted(self.elements, key=lambda x: x)[:3]
        )

    def test_do_all(self):
        class T(DoAll):
            def sort_key(self, chosen) -> int:
                return chosen

        inst = T()
        self.assertTrue(inst(self.elements) == sorted(self.elements, key=lambda x: x))

    def test_do_one_static(self):
        class T(DoOne):
            @staticmethod
            def sort_key(chosen) -> int:
                return chosen

        inst = T()
        self.assertTrue(inst(self.elements) == min(self.elements))

    def test_do_k_static(self):
        class T(DoK):
            @staticmethod
            def sort_key(chosen) -> int:
                return chosen

        inst = T()
        self.assertTrue(
            inst(3, self.elements) == sorted(self.elements, key=lambda x: x)[:3]
        )

    def test_do_all_static(self):
        class T(DoAll):
            @staticmethod
            def sort_key(chosen) -> int:
                return chosen

        inst = T()
        self.assertTrue(inst(self.elements) == sorted(self.elements, key=lambda x: x))

    # reverse tests
    def test_do_reverse(self):
        with self.assertRaises(InvalidControllerMethodError):

            class T(Do):
                reverse_sort = True

                def sort_key(self, chosen) -> int:
                    return chosen

    def test_do_one_reverse(self):
        class T(DoOne):
            reverse_sort = True

            def sort_key(self, chosen) -> int:
                return chosen

        inst = T()
        self.assertTrue(inst(self.elements) == max(self.elements))

    def test_do_k_reverse(self):
        class T(DoK):
            reverse_sort = True

            def sort_key(self, chosen) -> int:
                return chosen

        inst = T()
        self.assertTrue(
            inst(3, self.elements)
            == sorted(self.elements, key=lambda x: x, reverse=True)[:3]
        )

    def test_do_all_reverse(self):
        class T(DoAll):
            reverse_sort = True

            def sort_key(self, chosen) -> int:
                return chosen

        inst = T()
        self.assertTrue(
            inst(self.elements) == sorted(self.elements, key=lambda x: x, reverse=True)
        )

    def test_do_one_static_reverse(self):
        class T(DoOne):
            reverse_sort = True

            @staticmethod
            def sort_key(chosen) -> int:
                return chosen

        inst = T()
        self.assertTrue(inst(self.elements) == max(self.elements))

    def test_do_k_static_reverse(self):
        class T(DoK):
            reverse_sort = True

            @staticmethod
            def sort_key(chosen) -> int:
                return chosen

        inst = T()
        self.assertTrue(
            inst(3, self.elements)
            == sorted(self.elements, key=lambda x: x, reverse=True)[:3]
        )

    def test_do_all_static_reverse(self):
        class T(DoAll):
            reverse_sort = True

            @staticmethod
            def sort_key(chosen) -> int:
                return chosen

        inst = T()
        self.assertTrue(
            inst(self.elements) == sorted(self.elements, key=lambda x: x, reverse=True)
        )


class TestSortCmp(unittest.TestCase):

    def setUp(self):
        self.elements = [random.randint(0, 1000) for _ in range(10)]

    def test_do(self):
        with self.assertRaises(InvalidControllerMethodError):

            class T(Do):
                def sort_cmp(self, a, b) -> int:
                    return -1 if a < b else 1 if a > b else 0

    def test_do_one(self):
        class T(DoOne):
            def sort_cmp(self, a, b) -> int:
                return -1 if a < b else 1 if a > b else 0

        inst = T()
        self.assertTrue(inst(self.elements) == min(self.elements))

    def test_do_k(self):
        class T(DoK):
            def sort_cmp(self, a, b) -> int:
                return -1 if a < b else 1 if a > b else 0

        inst = T()
        self.assertTrue(
            inst(3, self.elements) == sorted(self.elements, key=lambda x: x)[:3]
        )

    def test_do_all(self):
        class T(DoAll):
            def sort_cmp(self, a, b) -> int:
                return -1 if a < b else 1 if a > b else 0

        inst = T()
        self.assertTrue(inst(self.elements) == sorted(self.elements, key=lambda x: x))

    def test_do_one_static(self):
        class T(DoOne):
            @staticmethod
            def sort_cmp(a, b) -> int:
                return -1 if a < b else 1 if a > b else 0

        inst = T()
        self.assertTrue(inst(self.elements) == min(self.elements))

    def test_do_k_static(self):
        class T(DoK):
            @staticmethod
            def sort_cmp(a, b) -> int:
                return -1 if a < b else 1 if a > b else 0

        inst = T()
        self.assertTrue(
            inst(3, self.elements) == sorted(self.elements, key=lambda x: x)[:3]
        )

    def test_do_all_static(self):
        class T(DoAll):
            @staticmethod
            def sort_cmp(a, b) -> int:
                return -1 if a < b else 1 if a > b else 0

        inst = T()
        self.assertTrue(inst(self.elements) == sorted(self.elements, key=lambda x: x))

    # reverse tests
    def test_do_reverse(self):
        with self.assertRaises(InvalidControllerMethodError):

            class T(Do):
                reverse_sort = True

                def sort_cmp(self, a, b) -> int:
                    return -1 if a < b else 1 if a > b else 0

    def test_do_one_reverse(self):
        class T(DoOne):
            reverse_sort = True

            def sort_cmp(self, a, b) -> int:
                return -1 if a < b else 1 if a > b else 0

        inst = T()
        self.assertTrue(inst(self.elements) == max(self.elements))

    def test_do_k_reverse(self):
        class T(DoK):
            reverse_sort = True

            def sort_cmp(self, a, b) -> int:
                return -1 if a < b else 1 if a > b else 0

        inst = T()
        self.assertTrue(
            inst(3, self.elements)
            == sorted(self.elements, key=lambda x: x, reverse=True)[:3]
        )

    def test_do_all_reverse(self):
        class T(DoAll):
            reverse_sort = True

            def sort_cmp(self, a, b) -> int:
                return -1 if a < b else 1 if a > b else 0

        inst = T()
        self.assertTrue(
            inst(self.elements) == sorted(self.elements, key=lambda x: x, reverse=True)
        )

    def test_do_one_static_reverse(self):
        class T(DoOne):
            reverse_sort = True

            @staticmethod
            def sort_cmp(a, b) -> int:
                return -1 if a < b else 1 if a > b else 0

        inst = T()
        self.assertTrue(inst(self.elements) == max(self.elements))

    def test_do_k_static_reverse(self):
        class T(DoK):
            reverse_sort = True

            @staticmethod
            def sort_cmp(a, b) -> int:
                return -1 if a < b else 1 if a > b else 0

        inst = T()
        self.assertTrue(
            inst(3, self.elements)
            == sorted(self.elements, key=lambda x: x, reverse=True)[:3]
        )

    def test_do_all_static_reverse(self):
        class T(DoAll):
            reverse_sort = True

            @staticmethod
            def sort_cmp(a, b) -> int:
                return -1 if a < b else 1 if a > b else 0

        inst = T()
        self.assertTrue(
            inst(self.elements) == sorted(self.elements, key=lambda x: x, reverse=True)
        )


class TestAction(unittest.TestCase):
    def setUp(self):
        self.elements = [random.randint(0, 1000) for _ in range(10)]

    def test_do_no_return(self):
        self.passed = False

        class T(Do):
            def action(do_self):
                self.passed = True

        inst = T()
        inst()
        self.assertTrue(self.passed)

    def test_do_one_no_return(self):
        self.passed = False
        self.count = 0

        class T(DoOne):
            def action(do_self, chosen):
                self.passed = True
                self.count += 1

        inst = T()
        inst(self.elements)
        self.assertTrue(self.passed)
        self.assertTrue(self.count == 1)

    def test_do_k_no_return(self):
        self.passed = False
        self.count = 0

        class T(DoK):
            def action(do_self, chosen):
                self.passed = True
                self.count += 1

        k = 5
        inst = T()
        inst(k, self.elements)
        self.assertTrue(self.passed)
        self.assertTrue(self.count == 5)

    def test_do_all_no_return(self):
        self.passed = False
        self.count = 0

        class T(DoAll):
            def action(do_self, chosen):
                self.passed = True
                self.count += 1

        inst = T()
        inst(self.elements)
        self.assertTrue(self.passed)
        self.assertTrue(self.count == len(self.elements))

    # with return
    def test_do_return(self):
        self.passed = False

        class T(Do):
            def action(do_self):
                self.passed = True
                return True

        inst = T()
        result = inst()
        self.assertTrue(self.passed)
        self.assertTrue(result)

    def test_do_one_return(self):
        self.passed = False
        self.count = 0

        class T(DoOne):
            def action(do_self, chosen):
                self.passed = True
                self.count += 1
                return chosen

        inst = T()
        result = inst(self.elements)
        self.assertTrue(self.passed)
        self.assertTrue(self.count == 1)
        self.assertTrue(result == self.elements[0])

    def test_do_k_return(self):
        self.passed = False
        self.count = 0

        class T(DoK):
            def action(do_self, chosen):
                self.passed = True
                self.count += 1
                return chosen

        k = 5
        inst = T()
        result = inst(k, self.elements)
        self.assertTrue(self.passed)
        self.assertTrue(self.count == k)
        self.assertTrue(result == self.elements[:k])

    def test_do_all_return(self):
        self.passed = False
        self.count = 0

        class T(DoAll):
            def action(do_self, chosen):
                self.passed = True
                self.count += 1
                return chosen

        inst = T()
        result = inst(self.elements)
        self.assertTrue(self.passed)
        self.assertTrue(self.count == len(self.elements))
        self.assertTrue(result == self.elements)

    # static tests
    def test_do_no_return_static(self):
        self.passed = False

        class T(Do):
            @staticmethod
            def action():
                self.passed = True

        inst = T()
        inst()
        self.assertTrue(self.passed)

    def test_do_one_no_return_static(self):
        self.passed = False
        self.count = 0

        class T(DoOne):
            @staticmethod
            def action(chosen):
                self.passed = True
                self.count += 1

        inst = T()
        inst(self.elements)
        self.assertTrue(self.passed)
        self.assertTrue(self.count == 1)

    def test_do_k_no_return_static(self):
        self.passed = False
        self.count = 0

        class T(DoK):
            @staticmethod
            def action(chosen):
                self.passed = True
                self.count += 1

        k = 5
        inst = T()
        inst(k, self.elements)
        self.assertTrue(self.passed)
        self.assertTrue(self.count == 5)

    def test_do_all_no_return_static(self):
        self.passed = False
        self.count = 0

        class T(DoAll):
            @staticmethod
            def action(chosen):
                self.passed = True
                self.count += 1

        inst = T()
        inst(self.elements)
        self.assertTrue(self.passed)
        self.assertTrue(self.count == len(self.elements))

    # with return
    def test_do_return_static(self):
        self.passed = False

        class T(Do):
            @staticmethod
            def action():
                self.passed = True
                return True

        inst = T()
        result = inst()
        self.assertTrue(self.passed)
        self.assertTrue(result)

    def test_do_one_return_static(self):
        self.passed = False
        self.count = 0

        class T(DoOne):
            @staticmethod
            def action(chosen):
                self.passed = True
                self.count += 1
                return chosen

        inst = T()
        result = inst(self.elements)
        self.assertTrue(self.passed)
        self.assertTrue(self.count == 1)
        self.assertTrue(result == self.elements[0])

    def test_do_k_return_static(self):
        self.passed = False
        self.count = 0

        class T(DoK):
            @staticmethod
            def action(chosen):
                self.passed = True
                self.count += 1
                return chosen

        k = 5
        inst = T()
        result = inst(k, self.elements)
        self.assertTrue(self.passed)
        self.assertTrue(self.count == k)
        self.assertTrue(result == self.elements[:k])

    def test_do_all_return_static(self):
        self.passed = False
        self.count = 0

        class T(DoAll):
            @staticmethod
            def action(chosen):
                self.passed = True
                self.count += 1
                return chosen

        inst = T()
        result = inst(self.elements)
        self.assertTrue(self.passed)
        self.assertTrue(self.count == len(self.elements))
        self.assertTrue(result == self.elements)


class TestFold(unittest.TestCase):
    def test_do(self):
        with self.assertRaises(InvalidControllerMethodError):

            class T(Do):
                def fold(self, results):
                    return 0

    def test_do_one(self):
        with self.assertRaises(InvalidControllerMethodError):

            class T(DoOne):
                def fold(self, results):
                    return 0

    def test_do_k(self):
        class T(DoK):
            def fold(self, results: List):
                return sum(results)

        inst = T()
        elements = [random.randint(0, 1000) for _ in range(10)]
        result = inst(5, elements)
        self.assertTrue(sum(elements[:5]) == result)

    def test_do_all(self):
        class T(DoAll):
            def fold(self, results: List):
                return sum(results)

        inst = T()
        elements = [random.randint(0, 1000) for _ in range(10)]
        result = inst(elements)
        self.assertTrue(sum(elements) == result)

    # staticmethods
    def test_do_static(self):
        with self.assertRaises(InvalidControllerMethodError):

            class T(Do):
                @staticmethod
                def fold(results):
                    return 0

    def test_do_one_static(self):
        with self.assertRaises(InvalidControllerMethodError):

            class T(DoOne):
                @staticmethod
                def fold(results):
                    return 0

    def test_do_k_static(self):
        class T(DoK):
            @staticmethod
            def fold(results: List):
                return sum(results)

        inst = T()
        elements = [random.randint(0, 1000) for _ in range(10)]
        result = inst(5, elements)
        self.assertTrue(sum(elements[:5]) == result)

    def test_do_all_static(self):
        class T(DoAll):
            @staticmethod
            def fold(results: List):
                return sum(results)

        inst = T()
        elements = [random.randint(0, 1000) for _ in range(10)]
        result = inst(elements)
        self.assertTrue(sum(elements) == result)


class TestPostController(unittest.TestCase):
    def test_do(self):
        class T(Do):
            passed = False

            def post_controller(self) -> None:
                self.passed = True

        inst = T()
        inst()
        self.assertTrue(inst.passed)

    def test_do_one(self):
        class T(DoOne):
            passed = False

            def post_controller(self) -> None:
                self.passed = True

        inst = T()
        inst([])
        self.assertTrue(inst.passed)

    def test_do_k(self):
        class T(DoK):
            passed = False

            def post_controller(self) -> None:
                self.passed = True

        inst = T()
        inst(0, [])
        self.assertTrue(inst.passed)

    def test_do_all(self):
        class T(DoAll):
            passed = False

            def post_controller(self) -> None:
                self.passed = True

        inst = T()
        inst([])
        self.assertTrue(inst.passed)

    # static method tests
    def test_do_static(self):
        class T(Do):
            @staticmethod
            def post_controller(arg1) -> None:
                arg1.value = True

        inst = T()
        arg = ArgWrapper(False)
        inst(arg)
        self.assertTrue(arg.value)

    def test_do_one_static(self):
        class T(DoOne):
            @staticmethod
            def post_controller(arg1) -> None:
                arg1.value = True

        inst = T()
        arg = ArgWrapper(False)
        inst([], arg)
        self.assertTrue(arg.value)

    def test_do_k_static(self):
        class T(DoK):
            @staticmethod
            def post_controller(arg1) -> None:
                arg1.value = True

        inst = T()
        arg = ArgWrapper(False)
        inst(0, [], arg)
        self.assertTrue(arg.value)

    def test_do_all_static(self):
        class T(DoAll):
            @staticmethod
            def post_controller(arg1) -> None:
                arg1.value = True

        inst = T()
        arg = ArgWrapper(False)
        inst([], arg)
        self.assertTrue(arg.value)


if __name__ == "__main__":
    unittest.main()



from setuptools import find_packages, setup

setup(
    name="metacontroller",
    version="0.1.1",
    author="Nathan Spencer",
    author_email="nathanss1997@gmail.com",
    description="A Python library for decoupling logic, modeling dynamic systems, and writing more declarative software.",
    long_description_content_type="text/markdown",
    url="https://github.com/nsspencer/MetaController",
    license="BSD 3-Clause",
    python_requires=">=3.8",
    packages=find_packages(exclude=["test"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    zip_safe=True,
    install_requires=[],  # no dependencies!
)



import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
import unittest

from metacontroller import Do

TEST_GLOBAL = False


class TestDoSmoke(unittest.TestCase):
    def test_basic_do(self):
        TEST_LOCAL = False

        class BaseDo(Do):
            def __init__(self) -> None:
                self.precontroller_passed = False
                self.postcontroller_passed = False
                self.action_passed = False

            def pre_controller(self) -> None:
                self.precontroller_passed = True

            def action(self):
                global TEST_GLOBAL
                nonlocal TEST_LOCAL
                self.action_passed = True
                TEST_LOCAL = True
                TEST_GLOBAL = True

            def post_controller(self) -> None:
                self.postcontroller_passed = True

        inst = BaseDo()

        # call the controller
        inst()
        self.assertTrue(inst.precontroller_passed)
        self.assertTrue(inst.postcontroller_passed)
        self.assertTrue(inst.action_passed)
        self.assertTrue(TEST_LOCAL)
        self.assertTrue(TEST_GLOBAL)

    def test_args_basic_do_return(test_self):

        class BaseDo(Do):
            def __init__(self) -> None:
                self.precontroller_passed = False
                self.postcontroller_passed = False
                self.action_passed = False

            def pre_controller(self, *args) -> None:
                self.precontroller_passed = True
                test_self.assertTrue(len(args) == 7)

            def action(self, arg0, **kwargs) -> int:
                self.action_passed = True
                test_self.assertTrue(arg0 == 0)
                test_self.assertTrue(kwargs.get("test", None) is not None)
                return arg0

            def post_controller(self, *args) -> None:
                self.postcontroller_passed = True
                test_self.assertTrue(len(args) == 7)

        inst = BaseDo()

        # call the controller
        result = inst(0, 1, 2, 3, 4, 5, 6, 7, test=1)
        test_self.assertTrue(inst.precontroller_passed)
        test_self.assertTrue(inst.postcontroller_passed)
        test_self.assertTrue(inst.action_passed)
        test_self.assertTrue(result == 0)

    def test_static(self):
        class StaticDo(Do):
            @staticmethod
            def action(arg0):
                return arg0 + 1

        inst = StaticDo()
        self.assertTrue(inst(0) == 1)


if __name__ == "__main__":
    unittest.main()



import random
from typing import Any

random.seed(0)
import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
import unittest

from metacontroller import DoOne


class TestDoOneMethodCombinations(unittest.TestCase):
    def setUp(test_self):
        test_self.elements = [random.randint(0, 1000) for _ in range(10)]

    def test_filter_sort_cmp(test_self):
        class T(DoOne):
            def filter(self, chosen) -> bool:
                return chosen % 2 == 0

            def sort_cmp(self, a, b):
                return -1 if a < b else 1 if a > b else 0

        inst = T()
        expected_result = sorted([i for i in test_self.elements if i % 2 == 0])[0]
        result = inst(test_self.elements)
        test_self.assertTrue(
            result == expected_result,
        )

    def test_filter_sort_key(test_self):
        class T(DoOne):
            def filter(self, chosen) -> bool:
                return chosen % 2 == 0

            def sort_key(self, chosen):
                return chosen

        inst = T()
        result = inst(test_self.elements)
        test_self.assertTrue(
            result == sorted(list(filter(lambda x: x % 2 == 0, test_self.elements)))[0]
        )

    def test_filter_action(test_self):
        class T(DoOne):
            def filter(self, chosen: Any) -> bool:
                return chosen % 2 == 0

            def action(self, chosen: Any) -> Any:
                return chosen + 1

        inst = T()
        result = inst(test_self.elements)
        test_self.assertTrue(
            result == [i + 1 for i in test_self.elements if i % 2 == 0][0]
        )

    def test_sort_cmp_action(test_self):
        class T(DoOne):
            def sort_cmp(self, a: Any, b: Any) -> int:
                return -1 if a < b else 1 if a > b else 0

            def action(self, chosen: Any) -> Any:
                return chosen + 1

        inst = T()
        result = inst(test_self.elements)
        expected_result = sorted([i + 1 for i in test_self.elements])[0]
        test_self.assertTrue(result == expected_result)

    def test_sort_key_action(test_self):
        class T(DoOne):
            def sort_key(self, chosen: Any):
                return chosen

            def action(self, chosen):
                return chosen + 1

        inst = T()
        result = inst(test_self.elements)
        test_self.assertTrue(result == [i + 1 for i in sorted(test_self.elements)][0])

    def test_filter_sort_cmp_action(test_self):
        class T(DoOne):
            def filter(self, chosen: Any) -> bool:
                return chosen % 2 == 0

            def sort_cmp(self, a: Any, b: Any) -> int:
                return -1 if a < b else 1 if a > b else 0

            def action(self, chosen):
                return chosen + 1

        inst = T()
        result = inst(test_self.elements)
        expected_results = sorted([i + 1 for i in test_self.elements if i % 2 == 0])[0]
        test_self.assertTrue(result == expected_results)

    def test_filter_sort_key_action(test_self):
        class T(DoOne):
            def filter(self, chosen: Any) -> bool:
                return chosen % 2 == 0

            def sort_key(self, chosen: Any):
                return chosen

            def action(self, chosen: Any) -> Any:
                return chosen + 1

        inst = T()
        result = inst(test_self.elements)
        expected_result = sorted([i + 1 for i in test_self.elements if i % 2 == 0])[0]
        test_self.assertTrue(result == expected_result)


if __name__ == "__main__":
    unittest.main()



import ast
import warnings
from typing import Any, Callable

from metacontroller.internal.method_invocation import MethodInvocation
from metacontroller.internal.namespace import (
    ACTION_METHOD_NAME,
    ACTION_RESULT_ASSIGNMENT_NAME,
    GENERATED_CALL_METHOD_NAME,
    POST_CONTROLLER_METHOD_NAME,
    PRE_CONTROLLER_METHOD_NAME,
)

from ._base import BaseControllerImplementation


class DoImplementation(BaseControllerImplementation):
    def __init__(self, cls, name, bases, attrs, stack_frame) -> None:
        super().__init__(
            cls,
            name,
            bases,
            attrs,
            stack_frame,
            filter_enabled=False,
            sort_key_enabled=False,
            sort_cmp_enabled=False,
            fold_enabled=False,
        )

    def validate(self) -> None:
        super().validate()
        if self.has_filter:
            warnings.warn(
                "Filter is not supported for Do controllers. It will be ignored."
            )

        if self.has_sort_cmp or self.has_sort_key:
            warnings.warn(
                "Sorting is not supported for Do controllers. It will be ignored."
            )

        if self.has_fold:
            warnings.warn(
                "Fold is not supported for Do controllers. It will be ignored."
            )

    def generate_call_method(self) -> Callable[..., Any]:
        body = []

        if self.has_pre_controller:
            pre_controller_call = MethodInvocation(
                self.pre_controller
            ).to_function_call(name=PRE_CONTROLLER_METHOD_NAME)
            body.append(ast.Expr(value=pre_controller_call))

        if self.has_action:
            result = ast.Assign(
                targets=[ast.Name(id=ACTION_RESULT_ASSIGNMENT_NAME, ctx=ast.Store())],
                value=MethodInvocation(self.action).to_function_call(
                    name=ACTION_METHOD_NAME
                ),
            )
            body.append(result)

        if self.has_post_controller:
            post_controller_call = MethodInvocation(
                self.post_controller
            ).to_function_call(name=POST_CONTROLLER_METHOD_NAME)
            body.append(ast.Expr(value=post_controller_call))

        if self.has_action:
            body.append(
                ast.Return(
                    value=ast.Name(id=ACTION_RESULT_ASSIGNMENT_NAME, ctx=ast.Load())
                )
            )

        args, saved_defaults = self.get_call_args(
            use_class_arg=True,
            use_k_arg=False,
            use_partition_arg=False,
            required_action_args=0,
        )
        call_fn = ast.FunctionDef(
            name=GENERATED_CALL_METHOD_NAME,
            args=args,
            body=body,
            decorator_list=[],
            type_params=[],
        )

        module = ast.fix_missing_locations(ast.Module(body=[call_fn], type_ignores=[]))
        return self.compile_call_method(module, saved_defaults)



import ast
import warnings
from functools import cmp_to_key
from heapq import nlargest, nsmallest
from itertools import islice
from typing import Any, Callable

from metacontroller.internal.exceptions import InvalidControllerMethodError
from metacontroller.internal.method_invocation import MethodInvocation
from metacontroller.internal.namespace import (
    ACTION_METHOD_NAME,
    ACTION_RESULT_ASSIGNMENT_NAME,
    CHOSEN_ARG_NAME,
    CLASS_ARG_NAME,
    FILTER_METHOD_NAME,
    FOLD_METHOD_NAME,
    GENERATED_CALL_METHOD_NAME,
    PARTITION_ARG_NAME,
    POST_CONTROLLER_METHOD_NAME,
    PRE_CONTROLLER_METHOD_NAME,
    SORT_CMP_METHOD_NAME,
    SORT_KEY_METHOD_NAME,
)

from ._base import BaseControllerImplementation


class DoOneImplementation(BaseControllerImplementation):
    def __init__(self, cls, name, bases, attrs, stack_frame) -> None:
        super().__init__(cls, name, bases, attrs, stack_frame, fold_enabled=False)

    def validate(self) -> None:
        super().validate()
        if self.has_sort_key and self.has_sort_cmp:
            err = f'DoOne controller "{self.name}" is invalid because both sort methods ("{SORT_KEY_METHOD_NAME}" and "{SORT_CMP_METHOD_NAME}") are defined.'
            err += f' You must define only one. Note that "{SORT_KEY_METHOD_NAME}" is more performant.'
            raise InvalidControllerMethodError(err)

        if self.has_fold:
            warnings.warn(
                f'DoOne does not support the "{FOLD_METHOD_NAME}" method, but one is defined in "{self.name}". It will be ignored.'
            )

        if self.has_filter:
            if len(self.filter.call_args) < 1:
                raise AttributeError(
                    f'"{FILTER_METHOD_NAME}" should be defined with at least 1 non-class argument (chosen), but 0 were given.'
                )

        if self.has_sort_key:
            if len(self.sort_key.call_args) < 1:
                raise AttributeError(
                    f'"{SORT_KEY_METHOD_NAME}" should be defined with at least 1 non-class argument (chosen), but 0 were given.'
                )

        if self.has_sort_cmp:
            if len(self.sort_cmp.call_args) < 2:
                raise AttributeError(
                    f'"{SORT_CMP_METHOD_NAME}" should be defined with at least 2 non-class arguments (a, b), but {len(self.sort_cmp.call_args)} were given.'
                )

        if self.has_action:
            if len(self.action.call_args) < 1:
                raise AttributeError(
                    f'"{ACTION_METHOD_NAME}" should be defined with at least 1 non-class argument (chosen), but 0 were given.'
                )

    def generate_call_method(self) -> Callable[..., Any]:
        body = []
        additional_globals = {}
        get_elements = ast.Name(id=PARTITION_ARG_NAME, ctx=ast.Load())

        if self.has_pre_controller:
            pre_controller_call = MethodInvocation(
                self.pre_controller
            ).to_function_call(name=PRE_CONTROLLER_METHOD_NAME)
            body.append(ast.Expr(value=pre_controller_call))

        if self.has_filter:
            if self.filter.num_call_parameters != 1:
                filter_fn = MethodInvocation(self.filter).to_lambda(
                    [self.filter.call_args[0]], name=FILTER_METHOD_NAME
                )
            else:
                filter_fn = ast.Attribute(
                    value=ast.Name(id=CLASS_ARG_NAME, ctx=ast.Load()),
                    attr=FILTER_METHOD_NAME,
                    ctx=ast.Load(),
                )
            get_elements = ast.Call(
                func=ast.Name(id="filter", ctx=ast.Load()),
                args=[filter_fn, get_elements],
                keywords=[],
            )

        if self.has_sort_key:
            if self.sort_key.num_call_parameters != 1:
                sort_fn_key = MethodInvocation(self.sort_key).to_lambda(
                    [self.sort_key.call_args[0]], name=SORT_KEY_METHOD_NAME
                )
            else:
                sort_fn_key = ast.Attribute(
                    value=ast.Name(id=CLASS_ARG_NAME, ctx=ast.Load()),
                    attr=SORT_KEY_METHOD_NAME,
                    ctx=ast.Load(),
                )

            if self.cls.reverse_sort:
                sort_fn = ast.Name(id="nlargest", ctx=ast.Load())
                additional_globals["nlargest"] = nlargest
            else:
                sort_fn = ast.Name(id="nsmallest", ctx=ast.Load())
                additional_globals["nsmallest"] = nsmallest

            get_elements = ast.Call(
                func=sort_fn,
                args=[ast.Constant(value=1, kind="int"), get_elements],
                keywords=[ast.keyword(arg="key", value=sort_fn_key)],
            )

        if self.has_sort_cmp:
            if self.sort_cmp.num_call_parameters != 2:
                sort_fn_key = MethodInvocation(self.sort_cmp).to_lambda(
                    self.sort_cmp.call_args[:2], name=SORT_CMP_METHOD_NAME
                )
            else:
                sort_fn_key = ast.Attribute(
                    value=ast.Name(id=CLASS_ARG_NAME, ctx=ast.Load()),
                    attr=SORT_CMP_METHOD_NAME,
                    ctx=ast.Load(),
                )

            if self.cls.reverse_sort:
                sort_fn = ast.Name(id="nlargest", ctx=ast.Load())
                additional_globals["nlargest"] = nlargest
            else:
                sort_fn = ast.Name(id="nsmallest", ctx=ast.Load())
                additional_globals["nsmallest"] = nsmallest

            get_elements = ast.Call(
                func=sort_fn,
                args=[ast.Constant(value=1, kind="int"), get_elements],
                keywords=[
                    ast.keyword(
                        arg="key",
                        value=ast.Call(
                            func=ast.Name(id="cmp_to_key", ctx=ast.Load()),
                            args=[sort_fn_key],
                            keywords=[],
                        ),
                    )
                ],
            )
            additional_globals["cmp_to_key"] = cmp_to_key

        if not self.has_sort_key and not self.has_sort_cmp:
            get_elements = ast.Call(
                func=ast.Name(id="islice", ctx=ast.Load()),
                args=[get_elements, ast.Constant(value=1, kind="int")],
                keywords=[],
            )
            additional_globals["islice"] = islice

            get_elements = ast.Call(
                func=ast.Name(id="list", ctx=ast.Load()),
                args=[get_elements],
                keywords=[],
            )

        # get the chosen element
        chosen_element = ast.Assign(
            targets=[ast.Name(id=CHOSEN_ARG_NAME, ctx=ast.Store())], value=get_elements
        )
        body.append(chosen_element)

        if self.has_action:
            action_invoke = MethodInvocation(self.action)
            action_args, action_keywords = action_invoke.get_call_args_and_keywords()

            # remove the original argument and replace it with the index 0 of the get_elements statement
            action_args.pop(0)
            action_args.insert(
                0,
                ast.Subscript(
                    value=ast.Name(id=CHOSEN_ARG_NAME, ctx=ast.Load()),
                    slice=ast.Index(value=ast.Constant(value=0)),
                    ctx=ast.Load(),
                ),
            )
            action_result = ast.Assign(
                targets=[ast.Name(id=ACTION_RESULT_ASSIGNMENT_NAME, ctx=ast.Store())],
                value=MethodInvocation(self.action).to_function_call(
                    action_args, action_keywords, name=ACTION_METHOD_NAME
                ),
            )
        else:
            action_result = ast.Assign(
                targets=[ast.Name(id=ACTION_RESULT_ASSIGNMENT_NAME, ctx=ast.Store())],
                value=ast.Subscript(
                    value=ast.Name(id=CHOSEN_ARG_NAME, ctx=ast.Load()),
                    slice=ast.Index(value=ast.Constant(value=0)),
                    ctx=ast.Load(),
                ),
            )

        # check if there is an element that we should act on
        if_check = ast.If(
            test=ast.Compare(
                left=ast.Call(
                    func=ast.Name(id="len", ctx=ast.Load()),
                    args=[ast.Name(id=CHOSEN_ARG_NAME, ctx=ast.Load())],
                    keywords=[],
                ),
                ops=[ast.NotEq()],
                comparators=[ast.Constant(value=0, kind="int")],
            ),
            body=[action_result],
            orelse=[],
        )

        result = ast.Assign(
            targets=[ast.Name(id=ACTION_RESULT_ASSIGNMENT_NAME, ctx=ast.Store())],
            value=ast.Constant(value=None, kind=None),
        )
        body.append(result)
        body.append(if_check)

        if self.has_post_controller:
            post_controller_call = MethodInvocation(
                self.post_controller
            ).to_function_call(name=POST_CONTROLLER_METHOD_NAME)
            body.append(ast.Expr(value=post_controller_call))

        if self.has_action:
            body.append(
                ast.Return(
                    value=ast.Name(id=ACTION_RESULT_ASSIGNMENT_NAME, ctx=ast.Load())
                )
            )
        else:
            body.append(
                ast.Return(
                    value=ast.Name(id=ACTION_RESULT_ASSIGNMENT_NAME, ctx=ast.Load())
                )
            )

        args, saved_defaults = self.get_call_args(
            use_class_arg=True,
            use_k_arg=False,
            use_partition_arg=True,
        )

        additional_globals.update(saved_defaults)
        call_fn = ast.FunctionDef(
            name=GENERATED_CALL_METHOD_NAME,
            args=args,
            body=body,
            decorator_list=[],
            type_params=[],
        )

        module = ast.fix_missing_locations(ast.Module(body=[call_fn], type_ignores=[]))
        return self.compile_call_method(module, additional_globals)
